import collections
import copy
from unittest import TestCase

from luqum.tree import (
    NONE_ITEM, Group, Word, Phrase, AndOperation, OrOperation, Proximity, SearchField,
    Boost, Fuzzy, Regex,
)
from luqum.visitor import (
    PathTrackingTransformer, PathTrackingVisitor, TreeTransformer, TreeVisitor,
)


class TreeVisitorTestCase(TestCase):

    class BasicVisitor(TreeVisitor):
        """Dummy visitor, simply yielding a list of nodes. """

        def generic_visit(self, node, context):
            yield node
            yield from super().generic_visit(node, context)

    class TrackingParentsVisitor(TreeVisitor):
        """Visitor, yielding nodes and parents."""

        def generic_visit(self, node, context):
            yield node, context.get("parents")
            yield from super().generic_visit(node, context)

    class MROVisitor(TreeVisitor):

        def visit_or_operation(self, node, context):
            yield "{} OR {}".format(*node.children)
            yield from super().generic_visit(node, context)

        def visit_base_operation(self, node, context):
            yield "{} BASE_OP {}".format(*node.children)
            yield from super().generic_visit(node, context)

        def visit_word(self, node, parents=[]):
            yield node.value

    def test_generic_visit(self):
        tree = AndOperation(Word("foo"), Word("bar"))
        visitor = TreeVisitor()
        nodes = visitor.visit(tree)
        self.assertEqual(nodes, [])
        # with a context for coverage…
        nodes = visitor.visit(tree, context={})
        self.assertEqual(nodes, [])

    def test_basic_traversal(self):
        tree = AndOperation(Word("foo"), Word("bar"))
        visitor = self.BasicVisitor()
        nodes = visitor.visit(tree)
        self.assertListEqual([tree, Word("foo"), Word("bar")], nodes)

    def test_parents_tracking(self):
        tree = AndOperation(Word("foo"), Proximity(Phrase('"bar"'), 2))
        visitor = self.TrackingParentsVisitor(track_parents=True)
        nodes = visitor.visit(tree)
        self.assertListEqual(
            [
                (tree, None),
                (Word("foo"), (tree,)),
                (Proximity(Phrase('"bar"'), degree=2), (tree,)),
                (Phrase('"bar"'), (tree, Proximity(Phrase('"bar"'), 2))),
            ],
            nodes,
        )

    def test_parents_tracking_no_tracking(self):
        tree = AndOperation(Word("foo"), Phrase('"bar"'))
        # no parents tracking !
        visitor = self.TrackingParentsVisitor()
        nodes = visitor.visit(tree)
        self.assertListEqual([(tree, None), (Word("foo"), None), (Phrase('"bar"'), None)], nodes)

    def test_mro(self):
        visitor = self.MROVisitor()

        tree = OrOperation(Word('a'), Word('b'))
        result = visitor.visit(tree)
        self.assertEqual(list(result), ['a OR b', 'a', 'b'])

        # AndOperation has no specific method,
        # but inherists BaseOperation, hence uses visit_base_operation
        tree = AndOperation(Word('a'), Word('b'))
        result = visitor.visit(tree)
        self.assertEqual(list(result), ['a BASE_OP b', 'a', 'b'])


class TreeTransformerTestCase(TestCase):

    class BasicTransformer(TreeTransformer):
        """
        Dummy transformer that simply turn any Word node's value into "lol"
        """
        def visit_word(self, node, context):
            yield Word(context.get("replacement", 'lol'))

        def visit_phrase(self, node, context):
            yield from []

        def visit_base_operation(self, node, context):
            new_node, = super().generic_visit(node, context)
            # if new_node has no operands, it's like a removal
            if len(new_node.children) == 0:
                return
            # if we have only one operands return it
            elif len(new_node.children) == 1:
                yield new_node.children[0]
            else:
                # normal return
                yield new_node

    class TrackingParentsTransformer(TreeTransformer):

        def visit_word(self, node, context):
            new_node, = self.generic_visit(node, context)
            if any(isinstance(p, SearchField) for p in context["new_parents"]):
                new_node.value = "lol"
            yield new_node

    class RaisingTreeTransformer(TreeTransformer):

        def generic_visit(self, node, context):
            yield node
            yield node

    class RaisingTreeTransformer2(TreeTransformer):

        def generic_visit(self, node, context):
            raise ValueError("Random error")

    def test_basic_traversal(self):
        tree = AndOperation(Word("foo"), Word("bar"))

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)
        self.assertEqual(new_tree, AndOperation(Word("lol"), Word("lol")))

    def test_context_value(self):
        tree = AndOperation(Word("foo"), Word("bar"))

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree, context={"replacement": "rotfl"})
        self.assertEqual(new_tree, AndOperation(Word("rotfl"), Word("rotfl")))

    def test_no_transform(self):
        tree = AndOperation(NONE_ITEM, NONE_ITEM)
        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)
        self.assertEqual(new_tree, tree)

    def test_one_word(self):
        tree = Word("foo")
        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)
        self.assertEqual(new_tree, Word("lol"))

    def test_tracking_parents(self):
        tree = OrOperation(Word("foo"), SearchField("test", Word("bar")))
        expected = OrOperation(Word("foo"), SearchField("test", Word("lol")))
        transformer = self.TrackingParentsTransformer(track_new_parents=True)
        new_tree = transformer.visit(tree)
        self.assertEqual(new_tree, expected)

    def test_removal(self):
        tree = AndOperation(
            OrOperation(Word("spam"), Word("ham")),
            AndOperation(Word("foo"), Phrase('"bar"')),
            AndOperation(Phrase('"baz"'), Phrase('"biz"')),
        )

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(
            new_tree,
            AndOperation(OrOperation(Word("lol"), Word("lol")), Word("lol")),
        )

    def test_silent_value_error(self):
        # in the case some attribute mislead the search for node do not raise
        tree = AndOperation(Word("a"), Word("b"))
        setattr(tree, "misleading1", ())
        setattr(tree, "misleading2", [])
        # hackishly patch __dict__ to be sure we have operands in right order for test
        tree.__dict__ = collections.OrderedDict(tree.__dict__)
        tree.__dict__['operands'] = tree.__dict__.pop('operands')  # operands are now last

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(new_tree, AndOperation(Word("lol"), Word("lol")))

    def test_repeating_expression(self):
        # non regression test
        tree = AndOperation(
            Group(OrOperation(Word('bar'), Word('foo'))),
            Group(OrOperation(Word('bar'), Word('foo'), Word('spam'))),
        )
        # basic transformer should not change tree
        same_tree = TreeTransformer().visit(copy.deepcopy(tree))
        self.assertEqual(same_tree, tree)

    def test_more_than_one_element_raises(self):
        tree = Word("foo")
        with self.assertRaises(ValueError) as raised:
            self.RaisingTreeTransformer().visit(tree)
        self.assertIn(
            "The visit of the tree should have produced exactly one element",
            str(raised.exception),
        )

    def test_value_error_pass_through(self):
        # raising a value error that is not related to unpacking passed through
        tree = Word("foo")
        with self.assertRaises(ValueError) as raised:
            self.RaisingTreeTransformer2().visit(tree)
        self.assertEqual("Random error", str(raised.exception))


class PathTrackingVisitorTestCase(TestCase):

    class TermPathVisitor(PathTrackingVisitor):

        def visit_term(self, node, context):
            yield (context["path"], node.value)

    @classmethod
    def setUpClass(cls):
        cls.visit = cls.TermPathVisitor().visit

    def test_visit_simple_term(self):
        paths = self.visit(Word("foo"))
        self.assertEqual(paths, [((), "foo")])

    def test_visit_complex(self):
        tree = AndOperation(
            Group(OrOperation(Word("foo"), Word("bar"), Boost(Fuzzy(Word("baz")), force=2))),
            Proximity(Phrase('"spam ham"')),
            SearchField("fizz", Regex("/fuzz/")),
        )
        paths = self.visit(tree)
        self.assertEqual(
            sorted(paths),
            [
                ((0, 0, 0), "foo"),
                ((0, 0, 1), "bar"),
                ((0, 0, 2, 0, 0), "baz"),
                ((1, 0), '"spam ham"'),
                ((2, 0), '/fuzz/'),
            ]
        )


class PathTrackingTransformerTestCase(TestCase):

    class TermPathTransformer(PathTrackingTransformer):

        def visit_term(self, node, context):
            path = '-'.join(str(i) for i in context['path'])
            quote = '"' if isinstance(node, Phrase) else "/" if isinstance(node, Regex) else ""
            value = node.value.strip(quote)
            new_node = node.clone_item(value=f"{quote}{value}@{path}{quote}")
            yield new_node

    @classmethod
    def setUpClass(cls):
        cls.transform = cls.TermPathTransformer().visit

    def test_visit_simple_term(self):
        tree = self.transform(Word("foo"))
        self.assertEqual(tree, Word("foo@"))

    def test_visit_complex(self):
        tree = AndOperation(
            Group(OrOperation(Word("foo"), Word("bar"), Boost(Fuzzy(Word("baz")), force=2))),
            Proximity(Phrase('"spam ham"')),
            SearchField("fizz", Regex("/fuzz/")),
        )
        transformed = self.transform(tree)
        expected = AndOperation(
            Group(OrOperation(
                Word("foo@0-0-0"),
                Word("bar@0-0-1"),
                Boost(Fuzzy(Word("baz@0-0-2-0-0")), force=2),
            )),
            Proximity(Phrase('"spam ham@1-0"')),
            SearchField("fizz", Regex("/fuzz@2-0/")),
        )
        self.assertEqual(transformed, expected)



from unittest import TestCase

from luqum.parser import parser
from luqum.tree import (Group, Word, AndOperation, OrOperation, BoolOperation,
                        UnknownOperation, Prohibit, Plus, From, To, Range, SearchField,
                        Boost)
from luqum.utils import UnknownOperationResolver, OpenRangeTransformer


class UnknownOperationResolverTestCase(TestCase):

    def test_and_resolution(self):
        tree = (
            UnknownOperation(
                Word("a"),
                Word("b"),
                OrOperation(Word("c"), Word("d"))))
        expected = (
            AndOperation(
                Word("a"),
                Word("b"),
                OrOperation(Word("c"), Word("d"))))
        resolver = UnknownOperationResolver(resolve_to=AndOperation)
        self.assertEqual(resolver(tree), expected)

    def test_or_resolution(self):
        tree = (
            UnknownOperation(
                Word("a"),
                Word("b"),
                AndOperation(Word("c"), Word("d"))))
        expected = (
            OrOperation(
                Word("a"),
                Word("b"),
                AndOperation(Word("c"), Word("d"))))
        resolver = UnknownOperationResolver(resolve_to=OrOperation)
        self.assertEqual(resolver(tree), expected)

    def test_lucene_resolution_simple(self):
        tree = (
            UnknownOperation(
                Word("a"),
                Word("b"),
                UnknownOperation(Word("c"), Word("d"))))
        expected = (
            AndOperation(
                Word("a"),
                Word("b"),
                AndOperation(Word("c"), Word("d"))))
        resolver = UnknownOperationResolver(resolve_to=None)
        self.assertEqual(resolver(tree), expected)

    def test_lucene_resolution_bool(self):
        tree = parser.parse("a b (+f +g) -(c d) +e")
        expected = (
            BoolOperation(
                Word("a"),
                Word("b"),
                Group(BoolOperation(Plus(Word("f")), Plus(Word("g")))),
                Prohibit(Group(BoolOperation(Word("c"), Word("d")))),
                Plus(Word('e'))))
        resolver = UnknownOperationResolver(resolve_to=BoolOperation)
        self.assertEqual(resolver(tree), expected)

    def test_lucene_resolution_last_op(self):
        tree = (
            OrOperation(
                Word("a"),
                Word("b"),
                UnknownOperation(Word("c"), Word("d")),
                AndOperation(
                    Word("e"),
                    UnknownOperation(Word("f"), Word("g"))),
                UnknownOperation(Word("i"), Word("j")),
                OrOperation(
                    Word("k"),
                    UnknownOperation(Word("l"), Word("m"))),
                UnknownOperation(Word("n"), Word("o"))))
        expected = (
            OrOperation(
                Word("a"),
                Word("b"),
                OrOperation(Word("c"), Word("d")),
                AndOperation(
                    Word("e"),
                    AndOperation(Word("f"), Word("g"))),
                AndOperation(Word("i"), Word("j")),
                OrOperation(
                    Word("k"),
                    OrOperation(Word("l"), Word("m"))),
                OrOperation(Word("n"), Word("o"))))
        resolver = UnknownOperationResolver(resolve_to=None)
        self.assertEqual(resolver(tree), expected)

    def test_lucene_resolution_last_op_with_group(self):
        tree = (
            OrOperation(
                Word("a"),
                Word("b"),
                Group(
                    AndOperation(
                        Word("c"),
                        UnknownOperation(Word("d"), Word("e")))),
                UnknownOperation(Word("f"), Word("g")),
                Group(
                    UnknownOperation(Word("h"), Word("i")))))
        expected = (
            OrOperation(
                Word("a"),
                Word("b"),
                Group(
                    AndOperation(
                        Word("c"),
                        AndOperation(Word("d"), Word("e")))),
                OrOperation(Word("f"), Word("g")),
                Group(
                    AndOperation(Word("h"), Word("i")))))
        resolver = UnknownOperationResolver(resolve_to=None)
        self.assertEqual(resolver(tree), expected)

    def test_resolve_to_verification(self):
        with self.assertRaises(ValueError):
            UnknownOperationResolver(resolve_to=object())

    def test_head_tail_pos(self):
        tree = parser.parse("\ra\nb (c\t (d e f)) ")
        resolver = UnknownOperationResolver(resolve_to=None)
        transformed = resolver(tree)
        self.assertEqual(str(transformed), "\ra\nAND b AND (c\t AND (d AND e AND f)) ")
        self.assertEqual(transformed.pos, tree.pos)
        self.assertEqual(transformed.size, tree.size)
        and_op, orig_op = transformed.children[2].children[0], tree.children[2].children[0]
        self.assertEqual(type(and_op), AndOperation)
        self.assertEqual(and_op.pos, orig_op.pos)
        self.assertEqual(and_op.size, orig_op.size)
        and_op, orig_op = and_op.children[1].children[0], orig_op.children[1].children[0]
        self.assertEqual(type(and_op), AndOperation)
        self.assertEqual(and_op.pos, orig_op.pos)
        self.assertEqual(and_op.size, orig_op.size)

        resolver = UnknownOperationResolver(resolve_to=OrOperation)
        transformed = resolver(tree)
        self.assertEqual(str(transformed), "\ra\nOR b OR (c\t OR (d OR e OR f)) ")


class OpenRangeTransformerTestCase(TestCase):
    def test_simple_resolution_from(self):
        tree = (
            From(Word("1"), True)
        )
        expected = (
            Range(Word("1", tail=" "), Word("*", head=" "), True, True)
        )
        for merge_ranges in (True, False):
            with self.subTest(merge_ranges=merge_ranges):
                resolver = OpenRangeTransformer(merge_ranges=merge_ranges)
                output = resolver(tree)
                self.assertEqual(output, expected)
                self.assertEqual(str(output), str(expected))

    def test_simple_resolution_to(self):
        tree = (
            To(Word("1"), False)
        )
        expected = (
            Range(Word("*", tail=" "), Word("1", head=" "), True, False)
        )
        for merge_ranges in (True, False):
            with self.subTest(merge_ranges=merge_ranges):
                resolver = OpenRangeTransformer(merge_ranges=merge_ranges)
                output = resolver(tree)
                self.assertEqual(output, expected)
                self.assertEqual(str(output), str(expected))

    def test_and_resolution(self):
        tree = (
            AndOperation(
                From(Word("1"), True),
                To(Word("2"), True),
            )
        )
        expected = (
            AndOperation(
                Range(Word("1", tail=" "), Word("2", head=" "), True, True)
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, expected)
        self.assertEqual(str(output), str(expected))

    def test_and_resolution_without_merge(self):
        tree = (
            AndOperation(
                From(Word("1"), True),
                To(Word("2"), True),
            )
        )
        expected = (
            AndOperation(
                Range(Word("1", tail=" "), Word("*", head=" "), True),
                Range(Word("*", tail=" "), Word("2", head=" "), True),
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=False)
        output = resolver(tree)
        self.assertEqual(output, expected)
        self.assertEqual(str(output), str(expected))

    def test_unjoined_resolution(self):
        tree = (
            AndOperation(
                From(Word("1"), False),
                From(Word("2"), True),
            )
        )
        expected = (
            AndOperation(
                Range(Word("1", tail=" "), Word("*", head=" "), False, True),
                Range(Word("2", tail=" "), Word("*", head=" "), True, True)
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, expected)
        self.assertEqual(str(output), str(expected))

    def test_normal_ranges_are_untouched(self):
        tree = (
            AndOperation(
                Range(Word("1"), Word("2"), True, True),
                Range(Word("*"), Word("*"), True, True),
                Range(Word("1"), Word("*"), True, True),
            )
        )
        for merge_ranges in (True, False):
            with self.subTest(merge_ranges=merge_ranges):
                resolver = OpenRangeTransformer(merge_ranges=merge_ranges)
                output = resolver(tree)
                self.assertEqual(output, tree)

    def test_first_range_is_merged(self):
        tree = (
            AndOperation(
                Range(Word("*"), Word("2"), True, True),
                Range(Word("*"), Word("*"), True, True),
                Range(Word("*"), Word("3"), True, True),
                Range(Word("1"), Word("*"), True, True),
                Range(Word("4"), Word("*"), True, True),
            )
        )
        expected = (
            AndOperation(
                Range(Word("1"), Word("2"), True, True),
                Range(Word("*"), Word("*"), True, True),
                Range(Word("4"), Word("3"), True, True),
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, expected)
        self.assertEqual(str(output), str(expected))

    def test_do_not_merge_unknown(self):
        tree = (
            UnknownOperation(
                Range(Word("1"), Word("*"), True, True),
                Range(Word("*"), Word("2"), True, True),
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, tree)

    def test_do_not_merge_searchfield(self):
        tree = (
            AndOperation(
                Range(Word("1"), Word("*"), True, True),
                SearchField("foo", Range(Word("*"), Word("2"), True, True))
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, tree)

    def test_do_not_merge_boosted(self):
        tree = (
            AndOperation(
                Boost(Range(Word("1"), Word("*"), True, True), 2),
                Boost(Range(Word("*"), Word("2"), True, True), 2),
            )
        )
        resolver = OpenRangeTransformer(merge_ranges=True)
        output = resolver(tree)
        self.assertEqual(output, tree)



from decimal import Decimal
from unittest import TestCase

from luqum.parser import parser
from luqum.tree import (
    SearchField, FieldGroup, Group, Item,
    Term, Word, Phrase, Regex, Proximity, Fuzzy, Boost, Range,
    NONE_ITEM, Not, AndOperation, OrOperation, Plus, Prohibit, UnknownOperation)


class TestTree(TestCase):

    def test_term_wildcard_true(self):
        self.assertTrue(Term("ba*").has_wildcard())
        self.assertTrue(Term("b*r").has_wildcard())
        self.assertTrue(Term("*ar").has_wildcard())
        self.assertTrue(Term("ba?").has_wildcard())
        self.assertTrue(Term("b?r").has_wildcard())
        self.assertTrue(Term("?ar").has_wildcard())
        self.assertTrue(Term("?a*").has_wildcard())
        self.assertTrue(Term(r"\?a*").has_wildcard())
        self.assertTrue(Term(r"?a\*").has_wildcard())
        self.assertTrue(Term("*").has_wildcard())
        self.assertTrue(Term("?").has_wildcard())
        # the \ itself can be escaped !
        self.assertTrue(Term(r"\\*\\?").has_wildcard())

    def test_term_wildcard_false(self):
        self.assertFalse(Term("bar").has_wildcard())
        self.assertFalse(Term(r"bar\*").has_wildcard())
        self.assertFalse(Term(r"b\?r\*").has_wildcard())

    def test_term_is_only_a_wildcard(self):
        self.assertTrue(Term('*').is_wildcard())
        self.assertFalse(Term('*o').is_wildcard())
        self.assertFalse(Term('b*').is_wildcard())
        self.assertFalse(Term('b*o').is_wildcard())
        self.assertFalse(Term('?').is_wildcard())

    def test_term_iter_wildcard(self):
        self.assertListEqual(
            list(Term(r"a?b\*or*and\?").iter_wildcards()),
            [((1, 2), "?"), ((7, 8), "*")],
            )
        self.assertListEqual(
            list(Term(r"\**\**").iter_wildcards()),
            [((2, 3), "*"), ((5, 6), "*")],
            )

    def test_term_split_wildcard(self):
        self.assertListEqual(
            Term(r"a??b\*or*and\?").split_wildcards(),
            ["a", "?", "", "?", r"b\*or", "*", r"and\?"],
            )
        self.assertListEqual(
            Term(r"\**\**").split_wildcards(),
            [r"\*", "*", r"\*", "*", ""],
            )

    def test_equality_proximty(self):
        """
        Regression test for a bug on approx equalities.

        .. todo:: Testing other tokens might be a good idea...
        """
        p1 = Proximity(term=Word('foo'), degree=5)
        p2 = Proximity(term=Word('bar'), degree=5)
        p3 = Proximity(term=Word('foo'), degree=5)
        p4 = Proximity(term=Word('foo'), degree=1)
        p5 = Proximity(term=Word('foo'), degree=None)

        self.assertNotEqual(p1, p2)
        self.assertEqual(p1, p3)
        self.assertNotEqual(p1, p4)
        self.assertEqual(p4, p5)

    def test_equality_fuzzy(self):
        f1 = Fuzzy(term=Word('foo'), degree=5)
        f2 = Fuzzy(term=Word('bar'), degree=5)
        f3 = Fuzzy(term=Word('foo'), degree=5)
        f4 = Fuzzy(term=Word('foo'), degree=.5)
        f5 = Fuzzy(term=Word('foo'), degree=None)

        self.assertNotEqual(f1, f2)
        self.assertEqual(f1, f3)
        self.assertNotEqual(f1, f4)
        self.assertEqual(f4, f5)

    def test_equality_boost(self):
        b1 = Boost(expr=Word('foo'), force=5)
        b2 = Boost(expr=Word('bar'), force=5)
        b3 = Boost(expr=Word('foo'), force=5)
        b4 = Boost(expr=Word('foo'), force=.5)

        self.assertNotEqual(b1, b2)
        self.assertEqual(b1, b3)
        self.assertNotEqual(b1, b4)

    def test_equality_range(self):
        r1 = Range(Word("20"), Word("40"), include_low=True, include_high=True)
        r2 = Range(Word("20"), Word("40"), include_low=True, include_high=True)
        self.assertEqual(r1, r2)

    def test_equality_range_different_terms(self):
        r1 = Range(Word("20"), Word("40"), include_low=True, include_high=True)
        self.assertNotEqual(r1, Range(Word("30"), Word("40"), include_low=True, include_high=True))
        self.assertNotEqual(r1, Range(Word("20"), Word("30"), include_low=True, include_high=True))

    def test_equality_range_different_bounds(self):
        """
        Regression test for a bug on range equalities.
        """
        r1 = Range(Word("20"), Word("40"), include_low=True, include_high=True)
        r2 = Range(Word("20"), Word("40"), include_low=False, include_high=True)
        r3 = Range(Word("20"), Word("40"), include_low=True, include_high=False)
        r4 = Range(Word("20"), Word("40"), include_low=False, include_high=False)
        self.assertNotEqual(r1, r2)
        self.assertNotEqual(r1, r3)
        self.assertNotEqual(r1, r4)
        self.assertNotEqual(r2, r3)
        self.assertNotEqual(r2, r4)
        self.assertNotEqual(r3, r4)

    def test_not_equality(self):
        # non regression test, adding a child should trigger non equality
        tree1 = OrOperation(Word("bar"))
        tree2 = OrOperation(Word("bar"), Word("foo"))
        self.assertNotEqual(tree1, tree2)


class SetChildrenTestCase(TestCase):

    def _test_set_children(self, item, children):
        item.children = children
        self.assertEqual(item.children, children)

    def test_set_children(self):
        test = self._test_set_children
        test(Word("foo"), [])
        test(Phrase('"foo"'), [])
        test(Regex("/foo/"), [])
        test(SearchField("foo", Word("bar")), [Word("baz")])
        test(Group(Word("foo")), [Word("foo")])
        test(FieldGroup(Word("foo")), [Word("foo")])
        test(Range(Word("20"), Word("30")), [Word("40"), Word("50")])
        test(Proximity(Word("foo")), [Word("foo")])
        test(Fuzzy(Word("foo")), [Word("foo")])
        test(Boost(Word("foo"), force=1), [Word("foo")])
        many_terms = tuple(Word(f"foo_{i}") for i in range(5))
        test(UnknownOperation(Word("foo"), Word("bar")), (Word("foo"), Word("bar")))
        test(UnknownOperation(*many_terms), many_terms)
        test(AndOperation(Word("foo"), Word("bar")), (Word("foo"), Word("bar")))
        test(AndOperation(*many_terms), many_terms)
        test(OrOperation(Word("foo"), Word("bar")), (Word("foo"), Word("bar")))
        test(OrOperation(*many_terms), many_terms)
        test(Plus(Word("foo")), [Word("foo")])
        test(Not(Word("foo")), [Word("foo")])
        test(Prohibit(Word("foo")), [Word("foo")])

    def _test_set_children_raises(self, item, children):
        with self.assertRaises(ValueError):
            item.children = children

    def test_set_children_raises(self):
        test = self._test_set_children_raises
        test(Word("foo"), [Word("foo")])
        test(Phrase('"foo"'), [Word("foo")])
        test(Regex("/foo/"), [Word("foo")])
        test(SearchField("foo", Word("bar")), [])
        test(SearchField("foo", Word("bar")), [Word("bar"), Word("baz")])
        test(Group(Word("foo")), [])
        test(Group(Word("foo")), [Word("foo"), Word("bar")])
        test(FieldGroup(Word("foo")), [])
        test(FieldGroup(Word("foo")), [Word("foo"), Word("bar")])
        test(Range(Word("20"), Word("30")), [])
        test(Range(Word("20"), Word("30")), [Word("20"), Word("30"), Word("40")])
        test(Proximity(Word("foo")), [])
        test(Proximity(Word("foo")), [Word("foo"), Word("bar")])
        test(Fuzzy(Word("foo")), [])
        test(Fuzzy(Word("foo")), [Word("foo"), Word("bar")])
        test(Boost(Word("foo"), force=1), [])
        test(Boost(Word("foo"), force=1), [Word("foo"), Word("bar")])
        test(Plus(Word("foo")), [])
        test(Plus(Word("foo")), [Word("foo"), Word("bar")])
        test(Not(Word("foo")), [])
        test(Not(Word("foo")), [Word("foo"), Word("bar")])
        test(Prohibit(Word("foo")), [])
        test(Prohibit(Word("foo")), [Word("foo"), Word("bar")])


class CloneTestCase(TestCase):

    def assert_equal_tail_head_pos(self, a, b):
        self.assertEqual(a.pos, b.pos)
        self.assertEqual(a.size, b.size)
        self.assertEqual(a.head, b.head)
        self.assertEqual(a.tail, b.tail)

    def test_word(self):
        orig = Word("foo", pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.value, copy.value)
        self.assertEqual(orig, copy)

    def test_clone_kwargs_overrides(self):
        orig = Word("foo", pos=3, head="\n", tail="\t")
        copy = orig.clone_item(value="bar")
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(copy.value, "bar")

    def test_phrase(self):
        orig = Phrase('"foo"', pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.value, copy.value)
        self.assertEqual(orig, copy)

    def test_phrase_to_word(self):
        orig = Phrase('"foo"', pos=3, head="\n", tail="\t")
        copy = orig._clone_item(cls=Word)  # making the phase a word
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.value, copy.value)
        self.assertTrue(isinstance(copy, Word))

    def test_regex(self):
        orig = Regex("/foo/", pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.value, copy.value)
        self.assertEqual(orig, copy)

    def test_search_field(self):
        orig = SearchField(name="foo", expr=Word("bar"), pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.name, copy.name)
        self.assertEqual(copy.expr, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def test_group(self):
        orig = Group(Word("bar"), pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(copy.expr, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def test_field_group(self):
        orig = FieldGroup(Word("bar"), pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(copy.expr, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def test_range(self):
        orig = Range(Word("foo"), Word("bar"), include_low=False, pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.include_low, copy.include_low)
        self.assertEqual(orig.include_high, copy.include_high)
        self.assertEqual(copy.low, NONE_ITEM)
        self.assertEqual(copy.high, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("foo"), Word("bar")]
        self.assertEqual(orig, copy)

    def test_proximity(self):
        orig = Proximity(Word("bar"), degree=3, pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.degree, copy.degree)
        self.assertEqual(copy.term, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def test_fuzzy(self):
        orig = Fuzzy(Word("bar"), degree=.3, pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.degree, copy.degree)
        self.assertEqual(copy.term, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def test_boost(self):
        orig = Boost(Word("bar"), force=3.2, pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(orig.force, copy.force)
        self.assertEqual(copy.expr, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("bar")]
        self.assertEqual(orig, copy)

    def _test_operation(self, cls):
        orig = cls(Word("foo"), Word("bar"), Word("baz"), pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(copy.operands, ())

        self.assertNotEqual(orig, copy)
        copy.children = [Word("foo"), Word("bar"), Word("baz")]
        self.assertEqual(orig, copy)

    def test_unknown_operation(self):
        self._test_operation(UnknownOperation)

    def test_and_operation(self):
        self._test_operation(AndOperation)

    def test_or_operation(self):
        self._test_operation(OrOperation)

    def _test_unary(self, cls):
        orig = cls(Word("foo"), pos=3, head="\n", tail="\t")
        copy = orig.clone_item()
        self.assert_equal_tail_head_pos(orig, copy)
        self.assertEqual(copy.a, NONE_ITEM)

        self.assertNotEqual(orig, copy)
        copy.children = [Word("foo")]
        self.assertEqual(orig, copy)

    def test_plus(self):
        self._test_unary(Plus)

    def test_not(self):
        self._test_unary(Not)

    def test_prohibit(self):
        self._test_unary(Prohibit)


class TestTreeSpan(TestCase):

    def test_simple(self):
        self.assertEqual(Item(pos=0, size=3).span(), (0, 3))
        self.assertEqual(Item(head="\r", tail="\t\t", pos=1, size=3).span(), (1, 4))
        self.assertEqual(Item(head="\r", tail="\t\t", pos=1, size=3).span(head_tail=True), (0, 6))

    def test_none(self):
        self.assertEqual(Item(pos=None, size=3).span(), (None, None))
        self.assertEqual(Item(pos=None, size=3).span(head_tail=True), (None, None))

    def test_integration(self):
        tree = parser.parse(" foo:bar OR baz OR ([20 TO 2000] AND more:(yee AND yii)) ")
        self.assertEqual(tree.span(), (0, 57))
        self.assertEqual(tree.span(head_tail=True), (0, 57))
        foo, baz, group = tree.children
        self.assertEqual(foo.span(), (1, 9))
        self.assertEqual(foo.span(head_tail=True), (0, 9))
        self.assertEqual(baz.span(), (12, 15))
        self.assertEqual(baz.span(head_tail=True), (11, 16))
        self.assertEqual(group.span(), (19, 56))
        self.assertEqual(group.span(head_tail=True), (18, 57))
        bar, = foo.children
        self.assertEqual(bar.span(), (5, 8))
        self.assertEqual(bar.span(head_tail=True), (5, 9))
        andop, = group.children
        self.assertEqual(andop.span(), (20, 55))
        self.assertEqual(andop.span(head_tail=True), (20, 55))
        range_, more = andop.children
        self.assertEqual(range_.span(), (20, 32))
        self.assertEqual(range_.span(head_tail=True), (20, 33))
        self.assertEqual(more.span(), (37, 55))
        self.assertEqual(more.span(head_tail=True), (36, 55))
        field_group, = more.children
        self.assertEqual(field_group.span(), (42, 55))
        self.assertEqual(field_group.span(head_tail=True), (42, 55))
        and_op2, = field_group.children
        yee, yii = and_op2.children
        self.assertEqual(yee.span(), (43, 46))
        self.assertEqual(yee.span(head_tail=True), (43, 47))
        self.assertEqual(yii.span(), (51, 54))
        self.assertEqual(yii.span(head_tail=True), (50, 54))


class TestPrint(TestCase):

    def test_unknown_operation(self):
        tree = UnknownOperation(Word("foo", tail=" "), Word("bar", tail=" "), Word("baz"))
        self.assertEqual(str(tree), "foo bar baz")

    def test_fuzzy(self):
        item = Fuzzy(Word("foo"), degree=None)
        self.assertEqual(str(item), "foo~")
        self.assertEqual(repr(item), "Fuzzy(Word('foo'), 0.5)")
        self.assertEqual(item.degree, Decimal(".5").normalize())
        item = Fuzzy(Word("foo"), degree=".5")
        self.assertEqual(str(item), "foo~0.5")
        item = Fuzzy(Word("foo"), degree=str(1/3))
        self.assertEqual(str(item), "foo~0.3333333333333333")
        # head tail
        item = Fuzzy(Word("foo", head="\t", tail="\n"), head="\r", tail="  ")
        self.assertEqual(str(item), "\tfoo\n~")
        self.assertEqual(item.__str__(head_tail=True), "\r\tfoo\n~  ")

    def test_proximity(self):
        item = Proximity(Word("foo"), degree=None)
        self.assertEqual(str(item), "foo~")
        self.assertEqual(repr(item), "Proximity(Word('foo'), 1)")
        self.assertEqual(item.degree, 1)
        item = Proximity(Word("foo"), degree="1")
        self.assertEqual(str(item), "foo~1")
        self.assertEqual(repr(item), "Proximity(Word('foo'), 1)")
        item = Proximity(Word("foo"), degree="4")
        self.assertEqual(str(item), "foo~4")
        self.assertEqual(repr(item), "Proximity(Word('foo'), 4)")
        # head tail
        item = Proximity(Word("foo", head="\t", tail="\n"), head="\r", tail="  ")
        self.assertEqual(str(item), "\tfoo\n~")
        self.assertEqual(item.__str__(head_tail=True), "\r\tfoo\n~  ")

    def test_boost(self):
        item = Boost(Word("foo"), force="3")
        self.assertEqual(str(item), "foo^3")
        self.assertEqual(repr(item), "Boost(Word('foo'), 3)")
        item = Boost(Word("foo"), force=str(1/3))
        self.assertEqual(str(item), "foo^0.3333333333333333")
        # head tail
        item = Boost(Word("foo", head="\t", tail="\n"), force=2, head="\r", tail="  ")
        self.assertEqual(str(item), "\tfoo\n^2")
        self.assertEqual(item.__str__(head_tail=True), "\r\tfoo\n^2  ")

    def test_none_item(self):
        self.assertEqual(str(NONE_ITEM), "")
        self.assertEqual(str(AndOperation(NONE_ITEM, NONE_ITEM)), "AND")



# -*- coding: utf-8 -*-
from unittest import TestCase

from luqum.naming import (
    auto_name, element_from_name, element_from_path, ExpressionMarker, get_name,
    HTMLMarker, matching_from_names, MatchingPropagator, set_name,
)
from luqum.parser import parser
from luqum.tree import (
    AndOperation, OrOperation, UnknownOperation, SearchField,
    Fuzzy, Proximity, Word, Phrase, Range, Regex, Group, FieldGroup,
    Plus, Not, Prohibit, Boost, Term,
)


def names_to_path(node, path=()):
    names = {}
    node_name = get_name(node)
    if node_name:
        names[node_name] = path
    for i, child in enumerate(node.children):
        names.update(names_to_path(child, path=path + (i,)))
    return names


def simple_naming(node, names=None, path=()):
    """utility to name a big tree, using the node class name or its content if it's a term

    If a name is repeated, it will add numbers. For example : `"or", "or2", "or3, …`.

    return dict: mapping names to path
    """
    if names is None:
        names = {}
    if isinstance(node, (Term)):
        node_name = node.value.strip('"').strip("/").lower()
    else:
        node_name = type(node).__name__.lower()
        if node_name.endswith("operation"):
            node_name = node_name[:-9]
    if node_name in names:
        node_name += str(1 + sum(1 for n in names if n.startswith(node_name)))
    set_name(node, node_name)
    names[node_name] = path
    for i, child in enumerate(node.children):
        simple_naming(child, names, path=path + (i,))
    return names


def paths_to_names(tree, paths):
    return {get_name(element_from_path(tree, path)) for path in paths}


class AutoNameTestCase(TestCase):

    def test_auto_name_one_term(self):
        tree = Word("test")
        names = auto_name(tree)
        self.assertEqual(get_name(tree), "a")
        self.assertEqual(names, {"a": ()})

        tree = Phrase('"test"')
        names = auto_name(tree)
        self.assertEqual(get_name(tree), "a")
        self.assertEqual(names, {"a": ()})

        tree = Range(Word("test"), Word("*"))
        names = auto_name(tree)
        self.assertEqual(get_name(tree), "a")
        self.assertEqual(names, {"a": ()})

        tree = Regex("/test/")
        names = auto_name(tree)
        self.assertEqual(get_name(tree), "a")
        self.assertEqual(names, {"a": ()})

    def test_auto_name_simple_op(self):
        for OpCls in AndOperation, OrOperation, UnknownOperation:
            with self.subTest("operation %r" % OpCls):
                tree = OpCls(
                    Word("test"),
                    Phrase('"test"'),
                )
                names = auto_name(tree)
                self.assertEqual(get_name(tree), None)
                self.assertEqual(get_name(tree.children[0]), "a")
                self.assertEqual(get_name(tree.children[1]), "b")
                self.assertEqual(names, {"a": (0,), "b": (1,)})

    def test_auto_name_nested(self):
        tree = AndOperation(
            OrOperation(
                SearchField("bar", Word("test")),
                AndOperation(
                    Proximity(Phrase('"test"'), 2),
                    SearchField("baz", Word("test")),
                ),
            ),
            Group(
                UnknownOperation(
                    Fuzzy(Word("test")),
                    Phrase('"test"'),
                ),
            ),
        )
        names = auto_name(tree)
        self.assertEqual(sorted(names.keys()), list("abcdefgh"))
        # and
        and1 = tree
        self.assertEqual(get_name(and1), None)
        # - or
        or1 = and1.children[0]
        self.assertEqual(get_name(or1), "a")
        self.assertEqual(names["a"], (0,))
        # -- search field
        sfield1 = or1.children[0]
        self.assertEqual(get_name(sfield1), "c")
        self.assertEqual(names["c"], (0, 0))
        self.assertEqual(get_name(sfield1.expr), None)
        # -- and
        and2 = or1.children[1]
        self.assertEqual(get_name(and2), "d")
        self.assertEqual(names["d"], (0, 1))
        # --- proximity phrase
        self.assertEqual(get_name(and2.children[0]), "e")
        self.assertEqual(names["e"], (0, 1, 0))
        self.assertEqual(get_name(and2.children[0].term), None)
        # --- search field
        sfield2 = and2.children[1]
        self.assertEqual(get_name(sfield2), "f")
        self.assertEqual(names["f"], (0, 1, 1))
        self.assertEqual(get_name(sfield2.expr), None)
        # - group
        group1 = and1.children[1]
        self.assertEqual(get_name(group1), "b")
        self.assertEqual(names["b"], (1,))
        # -- unknown op
        unknownop1 = group1.children[0]
        self.assertEqual(get_name(unknownop1), None)
        # --- fuzzy word
        self.assertEqual(get_name(unknownop1.children[0]), "g")
        self.assertEqual(names["g"], (1, 0, 0))
        self.assertEqual(get_name(unknownop1.children[0].term), None)
        # --- phrase
        self.assertEqual(get_name(unknownop1.children[1]), "h")
        self.assertEqual(names["h"], (1, 0, 1))


class UtilitiesTestCase(TestCase):

    def test_matching_from_name(self):
        names = {"a": (0,), "b": (1,), "c": (0, 0), "d": (0, 1), "e": (1, 0, 1)}
        self.assertEqual(
            matching_from_names([], names), (set(), {(0,), (1,), (0, 0), (0, 1), (1, 0, 1)})
        )
        self.assertEqual(
            matching_from_names(["a", "b"], names), ({(0,), (1,)}, {(0, 0), (0, 1), (1, 0, 1)})
        )
        self.assertEqual(
            matching_from_names(["a", "e"], names), ({(0,), (1, 0, 1)}, {(1,), (0, 0), (0, 1)})
        )
        self.assertEqual(
            matching_from_names(["c"], names), ({(0, 0)}, {(0,), (1,), (0, 1), (1, 0, 1)})
        )
        with self.assertRaises(KeyError):
            matching_from_names(["x"], names)

    def test_element_from_path(self):
        tree = AndOperation(
            OrOperation(
                SearchField("bar", Word("test")),
                Group(
                    AndOperation(
                        Proximity(Phrase('"test"'), 2),
                        SearchField("baz", Word("test")),
                        Fuzzy(Word("test")),
                        Phrase('"test"'),
                    ),
                ),
            ),
        )
        names = {"a": (), "b": (0, 1), "c": (0, 1, 0, 2), "d": (0, 1, 0, 2, 0), "e": (0, 1, 0, 3)}
        self.assertEqual(element_from_path(tree, ()), tree)
        self.assertEqual(element_from_name(tree, "a", names), tree)
        self.assertEqual(element_from_path(tree, (0, 1)), tree.children[0].children[1])
        self.assertEqual(element_from_name(tree, "b", names), tree.children[0].children[1])
        self.assertEqual(element_from_path(tree, (0, 1, 0, 2)), Fuzzy(Word("test")))
        self.assertEqual(element_from_name(tree, "c", names), Fuzzy(Word("test")))
        self.assertEqual(element_from_path(tree, (0, 1, 0, 2, 0)), Word("test"))
        self.assertEqual(element_from_name(tree, "d", names), Word("test"))
        self.assertEqual(element_from_path(tree, (0, 1, 0, 3)), Phrase('"test"'))
        self.assertEqual(element_from_name(tree, "e", names), Phrase('"test"'))
        with self.assertRaises(IndexError):
            element_from_path(tree, (1,))


class PropagateMatchingTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.propagate_matching = MatchingPropagator()

    def test_or_operation(self):
        tree = OrOperation(Word("foo"), Phrase('"bar"'), Word("baz"))
        all_paths = {(0,), (1,), (2,)}

        matching = set()
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, set())
        self.assertEqual(paths_ko, {(), (0,), (1,), (2,), })

        matching = {(2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(), (2, )})
        self.assertEqual(paths_ko, {(0,), (1,)})

        matching = {(0, ), (2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(), (0, ), (2,)})
        self.assertEqual(paths_ko, {(1,)})

        matching = {(0, ), (1, ), (2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(), (0,), (1,), (2, )})
        self.assertEqual(paths_ko, set())

    def test_and_operation(self):
        tree = AndOperation(Word("foo"), Phrase('"bar"'), Word("baz"))
        all_paths = {(0,), (1,), (2,)}

        matching = set()
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, set())
        self.assertEqual(paths_ko, {(), (0,), (1,), (2,), })

        matching = {(2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(2, )})
        self.assertEqual(paths_ko, {(), (0,), (1,)})

        matching = {(0, ), (2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(0, ), (2,)})
        self.assertEqual(paths_ko, {(), (1,)})

        matching = {(0, ), (1, ), (2, )}
        paths_ok, paths_ko = self.propagate_matching(tree, matching, all_paths - matching)
        self.assertEqual(paths_ok, {(), (0,), (1,), (2, )})
        self.assertEqual(paths_ko, set())

    def test_unknown_operation(self):
        tree = UnknownOperation(Word("foo"), Phrase('"bar"'), Word("baz"))

        tree_or = OrOperation(Word("foo"), Phrase('"bar"'), Word("baz"))
        tree_and = AndOperation(Word("foo"), Phrase('"bar"'), Word("baz"))
        propagate_or = self.propagate_matching
        propagate_and = MatchingPropagator(default_operation=AndOperation)
        all_paths = {(0,), (1,), (2,)}

        for matching in [set(), {(2, )}, {(0, ), (2, )}, {(0, ), (1, ), (2, )}]:
            self.assertEqual(
                propagate_or(tree, matching),
                self.propagate_matching(tree_or, matching, matching - all_paths),
            )
            self.assertEqual(
                propagate_and(tree, matching),
                self.propagate_matching(tree_and, matching, matching - all_paths),
            )

    def test_negation(self):
        for tree in [Prohibit(Word("foo")), Not(Word("foo"))]:
            with self.subTest("%r" % type(tree)):
                paths_ok, paths_ko = self.propagate_matching(tree, set(), {(0, )})
                self.assertEqual(paths_ok, {()})
                self.assertEqual(paths_ko, {(0,)})

                paths_ok, paths_ko = self.propagate_matching(tree, {(0, )}, set())
                self.assertEqual(paths_ok, {(0,)})
                self.assertEqual(paths_ko, {()})

    def test_nested_negation(self):
        for NegClass in (Prohibit, Not):
            with self.subTest("%r" % NegClass):
                tree = AndOperation(
                    NegClass(OrOperation(
                        NegClass(AndOperation(
                            NegClass(Word("a")),
                            Word("b"),
                        )),
                        Word("c"),
                    )),
                    Word("d"),
                )
                a, b, c, d = (0, 0, 0, 0, 0, 0), (0, 0, 0, 0, 1), (0, 0, 1), (1,)
                not_a, ab, not_ab = (0, 0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0)
                abc, not_abc, abcd = (0, 0), (0,), ()

                paths_ok, paths_ko = self.propagate_matching(tree, set(), {a, b, c, d})
                self.assertEqual(paths_ok, {not_a, not_ab, abc})
                self.assertEqual(paths_ko, {a, b, ab, c, not_abc, d, abcd})

                paths_ok, paths_ko = self.propagate_matching(tree, {b, d}, {a, c})
                self.assertEqual(paths_ok, {not_a, b, ab, not_abc, d, abcd})
                self.assertEqual(paths_ko, {a, not_ab, c, abc})

                paths_ok, paths_ko = self.propagate_matching(tree, {a, b, c}, {d})
                self.assertEqual(paths_ok, {a, b, not_ab, c, abc})
                self.assertEqual(paths_ko, {not_a, ab, not_abc, d, abcd})

                paths_ok, paths_ko = self.propagate_matching(tree, {a, b, c, d}, set())
                self.assertEqual(paths_ok, {a, b, not_ab, c, abc, d})
                self.assertEqual(paths_ko, {not_a, ab, not_abc, abcd})

    def test_single_element(self):
        for tree in [Word("a"), Phrase('"a"'), Regex("/a/")]:
            with self.subTest("%r" % type(tree)):
                paths_ok, paths_ko = self.propagate_matching(tree, set())
                self.assertEqual(paths_ok, set(), {()})
                self.assertEqual(paths_ko, {()})

                paths_ok, paths_ko = self.propagate_matching(tree, {()})
                self.assertEqual(paths_ok, {()}, set())
                self.assertEqual(paths_ko, set())

    def test_no_propagation(self):
        for tree in [Range(Word("a"), Word("b")), Fuzzy(Word("foo")), Proximity('"bar baz"', 2)]:
            with self.subTest("%r" % type(tree)):
                paths_ok, paths_ko = self.propagate_matching(tree, set(), {()})

                # no down propagation
                self.assertEqual(paths_ok, set())
                self.assertEqual(paths_ko, {()})

                paths_ok, paths_ko = self.propagate_matching(tree, {()}, set())
                self.assertEqual(paths_ok, {()})
                self.assertEqual(paths_ko, set())

    def test_simple_propagation(self):
        # propagation in nodes with only one children
        tree = Boost(Group(SearchField("foo", FieldGroup(Plus(Word("bar"))))), force=2)

        paths_ok, paths_ko = self.propagate_matching(tree, set(), {()})
        self.assertEqual(paths_ok, set())
        self.assertEqual(paths_ko, {(), (0,), (0, 0), (0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0, 0)})

        paths_ok, paths_ko = self.propagate_matching(tree, {()}, set())
        self.assertEqual(paths_ok, {(), (0,), (0, 0), (0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0, 0)})
        self.assertEqual(paths_ko, set())

    def test_combination(self):
        tree = AndOperation(
            OrOperation(
                SearchField(
                    "mine",
                    FieldGroup(
                        Plus(
                            AndOperation(
                                Word("foo"),
                                Regex("/fizz/"),
                            ),
                        ),
                    ),
                ),
                Boost(
                    Group(
                        AndOperation(
                            Phrase('"ham"'),
                            Word("spam"),
                            Prohibit(Fuzzy(Word("fuzz"))),
                        ),
                    ),
                    force=2,
                ),
            ),
            Not(
                OrOperation(
                    Word('"bar"'),
                    Word('"baz"'),
                ),
            ),
        )
        to_path = simple_naming(tree)

        paths_ok, paths_ko = self.propagate_matching(tree, set())
        self.assertEqual(
            paths_to_names(tree, paths_ok),
            {"prohibit", "not"},
        )
        self.assertEqual(
            paths_to_names(tree, paths_ko),
            {
                "and", "or", "searchfield", "fieldgroup", "plus", "and2", "foo", "fizz",
                "boost", "group", "and3", "ham", "spam", "fuzzy", "or2", "bar", "baz"
            },
        )

        # adding matching just enough positive expressions, so that complete expression matches
        paths_ok, paths_ko = self.propagate_matching(
            tree, {to_path["foo"], to_path["fizz"], to_path["ham"]},
        )
        self.assertEqual(
            paths_to_names(tree, paths_ok),
            {
                "and", "or", "searchfield", "fieldgroup", "plus", "and2", "foo", "fizz", "ham",
                "prohibit", "not"
            },
        )
        self.assertEqual(
            paths_to_names(tree, paths_ko),
            {"boost", "group", "and3", "spam", "fuzzy", "or2", "bar", "baz"},
        )

        # making everything match
        paths_ok, paths_ko = self.propagate_matching(
            tree, {to_path["foo"], to_path["fizz"], to_path["ham"], to_path["spam"]},
        )
        self.assertEqual(
            paths_to_names(tree, paths_ok),
            {"and", "or", "searchfield", "fieldgroup", "plus", "and2", "foo", "fizz", "ham",
             "prohibit", "boost", "group", "and3", "spam", "not"}
        )
        self.assertEqual(paths_to_names(tree, paths_ko), {"fuzzy", "or2", "bar", "baz"})

        # making everything match, but some negative expression
        paths_ok, paths_ko = self.propagate_matching(
            tree,
            {
                to_path["foo"], to_path["fizz"], to_path["ham"], to_path["spam"],
                to_path["fuzzy"], to_path["bar"],
            },
        )
        self.assertEqual(
            paths_to_names(tree, paths_ok),
            {
                "or", "searchfield", "fieldgroup", "plus", "and2",
                "foo", "fizz", "ham", "spam", "fuzzy", "or2", "bar",
            },
        )
        self.assertEqual(
            paths_to_names(tree, paths_ko),
            {
                "and", "boost", "group", "and3", "prohibit",
                "boost", "group", "and3", "not", "baz",
            },
        )


class HTMLMarkerTestCase(TestCase):

    mark_html = HTMLMarker()

    def test_single_element(self):
        ltree = parser.parse('"b"')
        out = self.mark_html(ltree, {()}, set())
        self.assertEqual(out, '<span class="ok">"b"</span>')
        out = self.mark_html(ltree, {()}, set(), parcimonious=False)
        self.assertEqual(out, '<span class="ok">"b"</span>')
        out = self.mark_html(ltree, set(), {()})
        self.assertEqual(out, '<span class="ko">"b"</span>')
        out = self.mark_html(ltree, set(), {()}, parcimonious=False)
        self.assertEqual(out, '<span class="ko">"b"</span>')

    def test_multiple_elements(self):
        ltree = parser.parse('(foo OR bar~2 OR baz^2) AND NOT spam')

        names = simple_naming(ltree)
        foo, bar, baz, spam = names["foo"], names["fuzzy"], names["boost"], names["spam"]
        or_, and_, not_ = names["or"], names["and"], names["not"]

        out = self.mark_html(ltree, {foo, bar, baz, or_, and_, not_}, {spam})
        self.assertEqual(
            out,
            '<span class="ok">(foo OR bar~2 OR baz^2) AND NOT<span class="ko"> spam</span></span>',
        )
        out = self.mark_html(ltree, {foo, bar, baz, or_, and_, not_}, {spam}, parcimonious=False)
        self.assertEqual(
            out,
            '<span class="ok">'
            '(<span class="ok"><span class="ok">foo </span>OR<span class="ok"> bar~2 </span>OR'
            '<span class="ok"> baz^2</span></span>) '
            'AND'
            '<span class="ok"> NOT<span class="ko"> spam</span></span>'
            '</span>',
        )

        out = self.mark_html(ltree, {not_}, {foo, bar, baz, or_, and_, spam})
        self.assertEqual(
            out,
            '<span class="ko">(foo OR bar~2 OR baz^2) AND'
            '<span class="ok"> NOT<span class="ko"> spam</span></span></span>',
        )

        # changing class name and element name
        mark = HTMLMarker(ok_class="success", ko_class="failure", element="li")
        out = mark(ltree, {not_}, {foo, bar, baz, or_, and_, spam})
        self.assertEqual(
            out,
            '<li class="failure">(foo OR bar~2 OR baz^2) AND'
            '<li class="success"> NOT<li class="failure"> spam</li></li></li>',
        )

    def test_expression_marker(self):
        # only for coverage !
        ltree = parser.parse("foo AND bar")
        mark = ExpressionMarker()
        out = mark(ltree, {(), (0,), (1,)}, {})
        self.assertEqual(out, ltree)



from unittest import TestCase

from luqum.pretty import Prettifier, prettify
from luqum.tree import (
    SearchField, FieldGroup, Group, Word, AndOperation, OrOperation, UnknownOperation)


class TestPrettify(TestCase):

    big_tree = AndOperation(
        Group(OrOperation(Word("baaaaaaaaaar"), Word("baaaaaaaaaaaaaz"))), Word("fooooooooooo"))
    fat_tree = AndOperation(
        SearchField(
            "subject",
            FieldGroup(
                OrOperation(
                    Word("fiiiiiiiiiiz"),
                    AndOperation(Word("baaaaaaaaaar"), Word("baaaaaaaaaaaaaz"))))),
        AndOperation(Word("fooooooooooo"), Word("wiiiiiiiiiz")))

    def test_one_liner(self):
        tree = AndOperation(Group(OrOperation(Word("bar"), Word("baz"))), Word("foo"))
        self.assertEqual(prettify(tree), "( bar OR baz ) AND foo")

    def test_with_unknown_op(self):
        prettify = Prettifier(indent=8, max_len=20)
        tree = UnknownOperation(
            Group(
                UnknownOperation(
                    Word("baaaaaaaaaar"),
                    Word("baaaaaaaaaaaaaz"))),
            Word("fooooooooooo"))
        self.assertEqual(
            "\n" + prettify(tree), """
(
        baaaaaaaaaar
        baaaaaaaaaaaaaz
)
fooooooooooo""")

    def test_with_unknown_op_nested(self):
        prettify = Prettifier(indent=8, max_len=20)
        tree = OrOperation(
            UnknownOperation(
                Word("baaaaaaaaaar"),
                Word("baaaaaaaaaaaaaz")),
            Word("fooooooooooo"))
        self.assertEqual(
            "\n" + prettify(tree), """
        baaaaaaaaaar
        baaaaaaaaaaaaaz
OR
fooooooooooo""")

    def test_small(self):
        prettify = Prettifier(indent=8, max_len=20)
        self.assertEqual(
            "\n" + prettify(self.big_tree), """
(
        baaaaaaaaaar
        OR
        baaaaaaaaaaaaaz
)
AND
fooooooooooo""")
        self.assertEqual(
            "\n" + prettify(self.fat_tree), """
subject: (
        fiiiiiiiiiiz
        OR
                baaaaaaaaaar
                AND
                baaaaaaaaaaaaaz
)
AND
fooooooooooo
AND
wiiiiiiiiiz""")

    def test_small_inline_ops(self):
        prettify = Prettifier(indent=8, max_len=20, inline_ops=True)
        self.assertEqual("\n" + prettify(self.big_tree), """
(
        baaaaaaaaaar OR
        baaaaaaaaaaaaaz ) AND
fooooooooooo""")
        self.assertEqual("\n" + prettify(self.fat_tree), """
subject: (
        fiiiiiiiiiiz OR
                baaaaaaaaaar AND
                baaaaaaaaaaaaaz ) AND
fooooooooooo AND
wiiiiiiiiiz""")

    def test_normal(self):
        prettify = Prettifier(indent=4, max_len=50)
        self.assertEqual("\n" + prettify(self.big_tree), """
(
    baaaaaaaaaar OR baaaaaaaaaaaaaz
)
AND
fooooooooooo""")
        self.assertEqual("\n" + prettify(self.fat_tree), """
subject: (
    fiiiiiiiiiiz
    OR
        baaaaaaaaaar AND baaaaaaaaaaaaaz
)
AND
fooooooooooo
AND
wiiiiiiiiiz""")

    def test_normal_inline_ops(self):
        prettify = Prettifier(indent=4, max_len=50, inline_ops=True)
        self.assertEqual("\n" + prettify(self.big_tree), """
(
    baaaaaaaaaar OR baaaaaaaaaaaaaz ) AND
fooooooooooo""")
        self.assertEqual("\n" + prettify(self.fat_tree), """
subject: (
    fiiiiiiiiiiz OR
        baaaaaaaaaar AND baaaaaaaaaaaaaz ) AND
fooooooooooo AND
wiiiiiiiiiz""")



from decimal import Decimal
from unittest import TestCase

import ply.lex as lex
import pytest

from luqum.exceptions import IllegalCharacterError, ParseSyntaxError
from luqum.parser import lexer, parser, parse
from luqum.tree import (
    SearchField, FieldGroup, Group,
    Word, Phrase, Regex, Proximity, Fuzzy, Boost, Range, From, To,
    Not, AndOperation, OrOperation, Plus, Prohibit, UnknownOperation)
from tests import alternative_lexer


class TestLexer(TestCase):
    """Test lexer
    """
    def test_basic(self):

        lexer.input(
            'subject:test desc:(house OR car)^3 AND "big garage"~2 dirt~0.3 OR foo:{a TO z*]')
        self.assertEqual(lexer.token().value, Word("subject"))
        self.assertEqual(lexer.token().type, "COLUMN")
        self.assertEqual(lexer.token().value, Word("test"))
        self.assertEqual(lexer.token().value, Word("desc"))
        self.assertEqual(lexer.token().type, "COLUMN")
        self.assertEqual(lexer.token().type, "LPAREN")
        self.assertEqual(lexer.token().value, Word("house"))
        self.assertEqual(lexer.token().type, "OR_OP")
        self.assertEqual(lexer.token().value, Word("car"))
        self.assertEqual(lexer.token().type, "RPAREN")
        t = lexer.token()
        self.assertEqual(t.type, "BOOST")
        self.assertEqual(t.value.value, "3")
        self.assertEqual(lexer.token().type, "AND_OP")
        self.assertEqual(lexer.token().value, Phrase('"big garage"'))
        t = lexer.token()
        self.assertEqual(t.type, "APPROX")
        self.assertEqual(t.value.value, "2")
        self.assertEqual(lexer.token().value, Word("dirt"))
        t = lexer.token()
        self.assertEqual(t.type, "APPROX")
        self.assertEqual(t.value.value, "0.3")
        self.assertEqual(lexer.token().type, "OR_OP")
        self.assertEqual(lexer.token().value, Word("foo"))
        self.assertEqual(lexer.token().type, "COLUMN")
        self.assertEqual(lexer.token().type, "LBRACKET")
        self.assertEqual(lexer.token().value, Word("a"))
        self.assertEqual(lexer.token().type, "TO")
        self.assertEqual(lexer.token().value, Word("z*"))
        self.assertEqual(lexer.token().type, "RBRACKET")
        self.assertEqual(lexer.token(), None)

    def test_accept_flavours(self):
        lexer.input('somedate:[now/d-1d+7H TO now/d+7H]')

        self.assertEqual(lexer.token().value, Word('somedate'))

        self.assertEqual(lexer.token().type, "COLUMN")
        self.assertEqual(lexer.token().type, "LBRACKET")

        self.assertEqual(lexer.token().value, Word("now/d-1d+7H"))
        self.assertEqual(lexer.token().type, "TO")
        self.assertEqual(lexer.token().value, Word("now/d+7H"))

        self.assertEqual(lexer.token().type, "RBRACKET")


class TestParser(TestCase):
    """Test base parser

    .. note:: we compare str(tree) before comparing tree, because it's more easy to debug
    """

    def test_simplest(self):
        tree = (
            AndOperation(
                Word("foo", tail=" "),
                Word("bar", head=" ")))
        parsed = parser.parse("foo AND bar")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_implicit_operations(self):
        tree = (
            UnknownOperation(
                Word("foo", tail=" "),
                Word("bar")))
        parsed = parser.parse("foo bar")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_simple_field(self):
        tree = (
            SearchField(
                "subject",
                Word("test")))
        parsed = parser.parse("subject:test")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_escaping_word(self):
        query = r'test\+\-\&\&\|\|\!\(\)\{\}\[\]\^\"\~\*\?\:\\test'
        tree = Word(query)
        unescaped = r'test+-&&||!(){}[]^"~*?:\test'
        parsed = parser.parse(query)
        self.assertEqual(str(parsed), query)
        self.assertEqual(parsed, tree)
        self.assertEqual(parsed.unescaped_value, unescaped)

    def test_escaping_word_first_letter(self):
        for letter in r'+-&|!(){}[]^"~*?:\\<>':
            with self.subTest("letter %s" % letter):
                query = r"\%stest" % letter
                tree = Word(query)
                unescaped = "%stest" % letter
                parsed = parser.parse(query)
                self.assertEqual(str(parsed), query)
                self.assertEqual(parsed, tree)
                self.assertEqual(parsed.unescaped_value, unescaped)

    def test_escaping_word_unbounded_equals(self):
        """This tests that <\\=foo correctly translates to <(=foo)"""
        query = r"<\=test"
        tree = To(Word("\\=test"), False)
        parsed = parser.parse(query)
        self.assertEqual(str(parsed), query)
        self.assertEqual(parsed, tree)
        self.assertEqual(parsed.a.unescaped_value, "=test")

    def test_escaping_phrase(self):
        query = r'"test \"ph\rase"'
        tree = Phrase(query)
        unescaped = '"test "phrase"'
        parsed = parser.parse(query)
        self.assertEqual(str(parsed), query)
        self.assertEqual(parsed, tree)
        self.assertEqual(parsed.unescaped_value, unescaped)

    def test_escaping_column(self):
        # non regression for issue #30
        query = r'ip:1000\:\:1000\:\:1/24'
        tree = SearchField('ip', Word(r'1000\:\:1000\:\:1/24'))
        parsed = parser.parse(query)
        self.assertEqual(parsed, tree)
        self.assertEqual(str(parsed), query)
        self.assertEqual(parsed.children[0].unescaped_value, "1000::1000::1/24")

    def test_escaping_single_column(self):
        # non regression for issue #30
        query = r'1000\:1000\:\:1/24'
        tree = Word(r'1000\:1000\:\:1/24')
        parsed = parser.parse(query)
        self.assertEqual(parsed, tree)
        self.assertEqual(str(parsed), query)
        self.assertEqual(parsed.unescaped_value, "1000:1000::1/24")

    def test_boost_default_to_one(self):
        query = "boost^ me^1"
        parsed = parser.parse(query)
        tree = UnknownOperation(Boost(Word("boost"), None), Boost(Word("me"), 1))
        self.assertEqual(parsed, tree)
        self.assertEqual(parsed.operands[0].force, 1)
        self.assertEqual(parsed.operands[0].implicit_force, True)
        self.assertEqual(parsed.operands[1].force, 1)
        self.assertEqual(parsed.operands[1].implicit_force, False)
        self.assertEqual(str(parsed), query)

    def test_field_with_number(self):
        # non regression for issue #10
        tree = (
            SearchField(
                "field_42",
                Word("42")))
        parsed = parser.parse("field_42:42")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_comma_can_be_a_term(self):
        # non regression for issue #79
        tree = UnknownOperation(Word("hi", tail=" "), Word(",", tail=" "), Word("bye"))
        parsed = parser.parse("hi , bye")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_minus(self):
        tree = (
            AndOperation(
                Prohibit(
                    Word("test", tail=" ")),
                Prohibit(
                    Word("foo", tail=" "), head=" "),
                Not(
                    Word("bar", head=" "), head=" ")))
        parsed = parser.parse("-test AND -foo AND NOT bar")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_plus(self):
        tree = (
            AndOperation(
                Plus(
                    Word("test", tail=" ")),
                Word("foo", head=" ", tail=" "),
                Plus(
                    Word("bar"), head=" ")))
        parsed = parser.parse("+test AND foo AND +bar")
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_phrase(self):
        tree = (
            AndOperation(
                Phrase('"a phrase (AND a complicated~ one)"', tail=" "),
                Phrase('"Another one"', head=" ")))
        parsed = parser.parse('"a phrase (AND a complicated~ one)" AND "Another one"')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_regex(self):
        tree = (
            AndOperation(
                Regex('/a regex (with some.*match+ing)?/', tail=" "),
                Regex('/Another one/', head=" ")))
        parsed = parser.parse('/a regex (with some.*match+ing)?/ AND /Another one/')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_approx(self):
        tree = (
            UnknownOperation(
                Proximity(
                    Phrase('"foo bar"'),
                    3,
                    tail=" "),
                Proximity(
                    Phrase('"foo baz"'),
                    None,
                    tail=" "),
                Fuzzy(
                    Word('baz'),
                    Decimal("0.3"),
                    tail=" "),
                Fuzzy(
                    Word('fou'),
                    None)))
        parsed = parser.parse('"foo bar"~3 "foo baz"~ baz~0.3 fou~')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_boost(self):
        tree = (
            UnknownOperation(
                Boost(
                    Phrase('"foo bar"'),
                    Decimal("3.0"),
                    tail=" "),
                Boost(
                    Group(
                        AndOperation(
                            Word('baz', tail=" "),
                            Word('bar', head=" "))),
                    Decimal("2.1"))))
        parsed = parser.parse('"foo bar"^3 (baz AND bar)^2.1')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_groups(self):
        tree = (
           OrOperation(
               Word('test', tail=" "),
               Group(
                   AndOperation(
                       SearchField(
                           "subject",
                           FieldGroup(
                               OrOperation(
                                   Word('foo', tail=" "),
                                   Word('bar', head=" "))),
                           tail=" "),
                       Word('baz', head=" ")),
                   head=" ")))
        parsed = parser.parse('test OR (subject:(foo OR bar) AND baz)')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_range(self):
        tree = (
            AndOperation(
                SearchField(
                    "foo",
                    Range(Word("10", tail=" "), Word("100", head=" "), True, True),
                    tail=" "),
                SearchField(
                    "bar",
                    Range(Word("a*", tail=" "), Word("*", head=" "), True, False),
                    head=" ")))
        parsed = parser.parse('foo:[10 TO 100] AND bar:[a* TO *}')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_flavours(self):
        tree = SearchField(
            "somedate",
            Range(Word("now/d-1d+7H", tail=" "), Word("now/d+7H", head=" "), True, True))
        parsed = parser.parse('somedate:[now/d-1d+7H TO now/d+7H]')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_combinations(self):
        # self.assertEqual(parser.parse("subject:test desc:(house OR car)").pval, "")
        tree = (
            UnknownOperation(
                SearchField(
                    "subject",
                    Word("test"),
                    tail=" "),
                AndOperation(
                    SearchField(
                        "desc",
                        FieldGroup(
                            OrOperation(
                                Word("house", tail=" "),
                                Word("car", head=" "))),
                        tail=" "),
                    Not(
                        Proximity(
                            Phrase('"approximatly this"'),
                            3,
                            head=" "),
                        head=" "))))
        parsed = parser.parse('subject:test desc:(house OR car) AND NOT "approximatly this"~3')

        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_from_to(self):
        """Tests that all unbounded ranges work correctly"""
        tree = (
            AndOperation(
                SearchField("foo", From(Word("10"), False), tail=" "),
                SearchField("bar", To(Word("11"), True), tail=" ", head=" "),
                SearchField("baz", From(Word("100"), True), tail=" ", head=" "),
                SearchField("fou", To(Phrase('"200"'), False), head=" ")
            )
        )
        parsed = parser.parse('foo:>10 AND bar:<=11 AND baz:>=100 AND fou:<"200"')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_from_to_fieldgroup(self):
        """Additional test for the use of combined open ranges (could have been a normal range)"""
        tree = (
            SearchField("foo", FieldGroup(UnknownOperation(
                From(Word("10"), True, tail=" "),
                To(Word("11"), False)
            )))
        )
        parsed = parser.parse('foo:(>=10 <11)')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_from_boost_plus_precedence(self):
        """Tests that boost, plus and from have correct precedence"""
        tree = Plus(Boost(From(Word("10"), True), 3))
        parsed = parser.parse('+>=10^3')
        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_precedence(self):
        """We test that unary + and - associate with the closest (unary) expression, and do not
        become associated with the entire ``2 AND 3`` expression

        Also, Boost must be closer than Plus: ``+1 AND +2^1`` must be ``+1 AND +(2^1)``
        """
        tree = (
            OrOperation(
                Word("1", tail=" "),
                AndOperation(
                    Plus(Boost(Fuzzy(Word("2"), 1), 1), tail=" "),
                    Word("3", head=" "),
                    head=" ",
                ),
            )
        )
        parsed = parser.parse('1 OR +2~1^1 AND 3')

        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_precedence_unknown(self):
        """Tests whether the implied unknown operation has the lowest associativity.
        This is independent of whether UnknownOperation resolves to AND or OR, as per
        evidence from Apache's Lucene parsing.

        Default operator: AND, parsed as: +1 +(+2 +3) +4
        Default operator: OR, parsed as:  1 (+2 +3) 4

        """
        tree = (
            UnknownOperation(
                Word("1", tail=" "),
                AndOperation(
                    Word("2", tail=" "),
                    Word("3", head=" "),
                    tail=" ",
                ),
                Word("4"),
            )
        )
        parsed = parser.parse('1 2 AND 3 4')

        self.assertEqual(str(parsed), str(tree))
        self.assertEqual(parsed, tree)

    def test_reserved_ok(self):
        """Test reserved word do not hurt in certain positions
        """
        tree = SearchField("foo", Word("TO"))
        parsed = parser.parse('foo:TO')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word("TO*"))
        parsed = parser.parse('foo:TO*')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word("NOT*"))
        parsed = parser.parse('foo:NOT*')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Phrase('"TO AND OR"'))
        parsed = parser.parse('foo:"TO AND OR"')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)

    def test_date_in_field(self):
        tree = SearchField("foo", Word("2015-12-19"))
        parsed = parser.parse('foo:2015-12-19')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word("2015-12-19T22:30"))
        parsed = parser.parse('foo:2015-12-19T22:30')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word("2015-12-19T22:30:45"))
        parsed = parser.parse('foo:2015-12-19T22:30:45')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word("2015-12-19T22:30:45.234Z"))
        parsed = parser.parse('foo:2015-12-19T22:30:45.234Z')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)

    def test_datemath_in_field(self):
        tree = SearchField("foo", Word(r"2015-12-19||+2\d"))
        parsed = parser.parse(r'foo:2015-12-19||+2\d')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)
        tree = SearchField("foo", Word(r"now+2h+20m\h"))
        parsed = parser.parse(r'foo:now+2h+20m\h')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)

    def test_date_in_range(self):
        # juste one funky expression
        tree = SearchField(
            "foo",
            Range(Word(r"2015-12-19||+2\d", tail=" "), Word(r"now+3d+12h\h", head=" ")))
        parsed = parser.parse(r'foo:[2015-12-19||+2\d TO now+3d+12h\h]')
        self.assertEqual(str(tree), str(parsed))
        self.assertEqual(tree, parsed)

    def test_reserved_ko(self):
        """Test reserved word hurt as they hurt lucene
        """
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('foo:NOT')
        self.assertTrue(
            str(raised.exception).startswith("Syntax error in input : unexpected end of expr"))
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('foo:AND')
        self.assertEqual(
            str(raised.exception),
            "Syntax error in input : unexpected  'AND' at position 4!",
        )
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('foo:OR')
        self.assertEqual(
            str(raised.exception),
            "Syntax error in input : unexpected  'OR' at position 4!",
        )
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('OR')
        self.assertEqual(
            str(raised.exception),
            "Syntax error in input : unexpected  'OR' at position 0!",
        )
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('AND')
        self.assertEqual(
            str(raised.exception),
            "Syntax error in input : unexpected  'AND' at position 0!",
        )

    def test_parse_error_on_unmatched_parenthesis(self):
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('((foo bar) ')
        self.assertTrue(
            str(raised.exception).startswith("Syntax error in input : unexpected end of expr"))

    def test_parse_error_on_unmatched_bracket(self):
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('[foo TO bar')
        self.assertTrue(
            str(raised.exception).startswith("Syntax error in input : unexpected end of expr"))

    def test_parse_error_on_range(self):
        with self.assertRaises(ParseSyntaxError) as raised:
            parser.parse('[foo TO ]')
        self.assertEqual(
            str(raised.exception),
            "Syntax error in input : unexpected  ']' at position 8!",
        )

    def test_illegal_character_exception(self):
        with self.assertRaises(IllegalCharacterError) as raised:
            parser.parse('\\')
        self.assertEqual(
            str(raised.exception),
            "Illegal character '\\' at position 0",
        )

    def test_negative_values_in_ranges(self):
        parsed = parser.parse("[-1 TO 5]")
        self.assertEqual(str(parsed), "[-1 TO 5]")

        parsed = parser.parse("[-10 TO -1]")
        self.assertEqual(str(parsed), "[-10 TO -1]")

        # semantically incorrect but correct from the parser's perspective
        parsed = parser.parse("[5 TO -1]")
        self.assertEqual(str(parsed), "[5 TO -1]")


def test_lex_global_state():
    """
    Last Lexer is used globally by default by the parser. If another library
    creates another lexer, it should not impact luqum.

    More info: [Multiple Parsers and
    Lexers](http://www.dabeaz.com/ply/ply.html#ply_nn37)
    """
    qs = '(title:"foo bar" AND body:"quick fox")'

    lex.lex(module=alternative_lexer)

    with pytest.raises(ParseSyntaxError):
        parser.parse(qs, lexer=lex.lexer)

    # if there is a "luqum.exceptions.ParseSyntaxError",
    # the wrong lexer was used.
    parse(qs)
    parser.parse(qs)



# -*- coding: utf-8 -*-
"""
.. todo:: split this file in multiple file: tree, lexer, parser
"""
import collections
import copy
from unittest import TestCase

from luqum.tree import Group, Word, Phrase, AndOperation, OrOperation
from luqum.deprecated_utils import LuceneTreeVisitor, LuceneTreeTransformer, LuceneTreeVisitorV2


class TreeVisitorTestCase(TestCase):

    class BasicVisitor(LuceneTreeVisitor):
        """ Dummy visitor, simply yielding a list of nodes. """
        def generic_visit(self, node, parents):
            yield node

    class MROVisitor(LuceneTreeVisitor):

        def visit_or_operation(self, node, parents=[]):
            return ["{} OR {}".format(*node.children)]

        def visit_base_operation(self, node, parents=[]):
            return ["{} BASE_OP {}".format(*node.children)]

        def visit_word(self, node, parents=[]):
            return [node.value]

    def test_generic_visit(self):
        tree = (
            AndOperation(
                Word("foo"),
                Word("bar")))

        visitor = LuceneTreeVisitor()
        nodes = list(visitor.visit(tree))
        self.assertEqual(nodes, [])

    def test_basic_traversal(self):
        tree = (
            AndOperation(
                Word("foo"),
                Word("bar")))

        visitor = self.BasicVisitor()
        nodes = list(visitor.visit(tree))

        self.assertListEqual(
            [AndOperation(Word('foo'), Word('bar')), Word('foo'), Word('bar')],
            nodes)

    def test_mro(self):
        visitor = self.MROVisitor()

        tree = OrOperation(Word('a'), Word('b'))
        result = visitor.visit(tree)
        self.assertEqual(list(result), ['a OR b', 'a', 'b'])

        tree = AndOperation(Word('a'), Word('b'))
        result = visitor.visit(tree)
        self.assertEqual(list(result), ['a BASE_OP b', 'a', 'b'])


class TreeTransformerTestCase(TestCase):

    class BasicTransformer(LuceneTreeTransformer):
        """
        Dummy transformer that simply turn any Word node's value into "lol"
        """
        def visit_word(self, node, parent):
            return Word('lol')

        def visit_phrase(self, node, parent):
            return None

    class OrListOperation(OrOperation):
        """Dummy operation having list operands instead of tuple
        """
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.operands = list(self.operands)

    def test_basic_traversal(self):
        tree = (
            AndOperation(
                Word("foo"),
                Word("bar")))

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(
            new_tree,
            (AndOperation(
                Word("lol"),
                Word("lol"))))

    def test_no_transform(self):
        tree = AndOperation()
        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)
        self.assertEqual(
            new_tree,
            AndOperation())

    def test_one_word(self):
        tree = Word("foo")
        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)
        self.assertEqual(
            new_tree,
            Word("lol"))

    def test_removal(self):
        tree = (
            AndOperation(
                AndOperation(
                    Word("foo"),
                    Phrase('"bar"')),
                AndOperation(
                    Phrase('"baz"'),
                    Phrase('"biz"'))))

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(
            new_tree,
            (AndOperation(
                AndOperation(Word("lol")),
                AndOperation())))

    def test_operands_list(self):
        OrListOperation = self.OrListOperation
        tree = (
            OrListOperation(
                OrListOperation(
                    Word("foo"),
                    Phrase('"bar"')),
                OrListOperation(
                    Phrase('"baz"'))))

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(
            new_tree,
            (OrListOperation(
                OrListOperation(Word("lol")),
                OrListOperation())))

    def test_silent_value_error(self):
        # in the case some attribute mislead the search for node do not raise
        tree = AndOperation(Word("a"), Word("b"))
        setattr(tree, "misleading1", ())
        setattr(tree, "misleading2", [])
        # hackishly patch __dict__ to be sure we have operands in right order for test
        tree.__dict__ = collections.OrderedDict(tree.__dict__)
        tree.__dict__['operands'] = tree.__dict__.pop('operands')  # operands are now last

        transformer = self.BasicTransformer()
        new_tree = transformer.visit(tree)

        self.assertEqual(
            new_tree,
            AndOperation(Word("lol"), Word("lol")))

    def test_repeating_expression(self):
        # non regression test
        tree = AndOperation(
            Group(OrOperation(Word('bar'), Word('foo'))),
            Group(OrOperation(Word('bar'), Word('foo'), Word('spam'))),
        )
        # basic transformer should not change tree
        same_tree = LuceneTreeTransformer().visit(copy.deepcopy(tree))
        self.assertEqual(same_tree, tree)


class TreeVisitorV2TestCase(TestCase):

    class BasicVisitor(LuceneTreeVisitorV2):
        """ Dummy visitor, simply yielding a list of nodes. """
        def generic_visit(self, node, parents, context):
            yield node
            for c in node.children:
                yield from self.visit(c, parents + [node], context)

    class MROVisitor(LuceneTreeVisitorV2):

        def visit_or_operation(self, node, parents=[], context=None):
            return "{} OR {}".format(*[self.visit(c) for c in node.children])

        def visit_base_operation(self, node, parents=[], context=None):
            return "{} BASE_OP {}".format(*[self.visit(c) for c in node.children])

        def visit_word(self, node, parents=[], context=None):
            return node.value

    def test_basic_traversal(self):
        tree = (
            AndOperation(
                Word("foo"),
                Word("bar")))

        visitor = self.BasicVisitor()
        nodes = list(visitor.visit(tree))

        self.assertListEqual(
            [AndOperation(Word('foo'), Word('bar')), Word('foo'), Word('bar')],
            nodes)

    def test_mro(self):
        visitor = self.MROVisitor()

        tree = OrOperation(Word('a'), Word('b'))
        result = visitor.visit(tree)
        self.assertEqual(result, 'a OR b')

        tree = OrOperation(AndOperation(Word('a'), Word('b')), Word('c'))
        result = visitor.visit(tree)
        self.assertEqual(result, 'a BASE_OP b OR c')

    def test_generic_visit_fails_by_default(self):
        visitor = self.MROVisitor()
        with self.assertRaises(AttributeError):
            visitor.visit(Phrase('"test"'))



# -*- coding: utf-8 -*-
"""Various utilities for dealing with syntax trees.

Include base classes to implement a visitor pattern.

"""
# utils pending deprecation
from . import visitor
from .deprecated_utils import (  # noqa: F401
    LuceneTreeTransformer, LuceneTreeVisitor, LuceneTreeVisitorV2)
from .tree import AndOperation, BaseOperation, OrOperation, BoolOperation, Range, Word


class UnknownOperationResolver(visitor.TreeTransformer):
    """Transform the UnknownOperation to OR or AND
    """

    VALID_OPERATIONS = frozenset([None, AndOperation, OrOperation, BoolOperation])
    DEFAULT_OPERATION = AndOperation

    def __init__(self, resolve_to=None, add_head=" "):
        """Initialize a new resolver

        :param resolve_to: must be either None, OrOperation, AndOperation, BoolOperation.

          for the latter three the UnknownOperation is replaced by specified operation.

          if it is None, we use the last operation encountered, as would Lucene do
        """
        if resolve_to not in self.VALID_OPERATIONS:
            raise ValueError("%r is not a valid value for resolve_to" % resolve_to)
        self.resolve_to = resolve_to
        self.add_head = add_head
        super().__init__(track_parents=True)

    def _last_operation(self, context):
        return context.setdefault("last_operation", {})

    def _first_nonop_parent(self, parents):
        for parent in parents:
            if not isinstance(parent, BaseOperation):
                return id(parent)  # use id() because parent might not be hashable
        else:
            return None

    def _track_last_op(self, node, context):
        if self.resolve_to is None:
            # next unknow operation at same level should resolve to my type of operation
            # so track it
            parent = self._first_nonop_parent(context.get("parents", []))
            self._last_operation(context)[parent] = type(node)

    def _get_last_op(self, node, context):
        # search for last operation
        parent = self._first_nonop_parent(context.get("parents", []))
        return self._last_operation(context).get(parent, self.DEFAULT_OPERATION)

    def visit_or_operation(self, node, context):
        self._track_last_op(node, context)
        yield from self.generic_visit(node, context)

    def visit_and_operation(self, node, context):
        self._track_last_op(node, context)
        yield from self.generic_visit(node, context)

    def visit_unknown_operation(self, node, context):
        # resolve
        if self.resolve_to is None:
            operation = self._get_last_op(node, context)
        else:
            operation = self.resolve_to
        new_node = operation(pos=node.pos, size=node.size, head=node.head, tail=node.tail)
        new_node.children = self.clone_children(node, new_node, context)
        # add head to children but the first to separate element from operation (x y --> x AND y)
        for child in new_node.children[1:]:
            child.head = self.add_head + child.head
        yield new_node

    def __call__(self, tree):
        return self.visit(tree)


class OpenRangeTransformer(visitor.TreeTransformer):
    """Transforms open ranges to normal Range objects, i.e.

    ::

        >=foo            ->   [foo TO *]
        <bar             ->   [* TO bar}

    When *merge_ranges* is set, this also merges open ranges in AND clauses::

        >foo AND <=bar              ->   {foo TO bar]
        [foo TO *] AND [* TO bar]   ->   [foo TO bar]

    The merging of open ranges is performed by collecting all open ranges, and merging any
    open range into any previously collected open range with an open bound. In other words, any
    open bounds are always merged into the first suitable. Additionally, matching is always done
    from left-to-right, so that this holds::

        [a TO *] AND [b TO *] AND [* TO y] AND [* TO z]   ->   [a TO y] AND [b TO z]

    Open ranges in OR and unknown clauses are not adjusted. Use :cls:`UnknownOperationResolver`
    to make sure that unknown operations are resolved first.

    Ranges with none of the bounds set are left unadjusted.

    Additionally, the ranges must be direct siblings of the same parent.
    Ranges such as ``[foo TO *]^2 AND [* TO bar]^2``
    are therefore not merged (though ``([foo TO *] AND [* TO bar])^2`` would).

    Also not that it does not check ranges bound ordering
    so that you might end-up with incoherent ranges like [* TO 3] AND [4 TO *] -> [4 TO 3]
    """

    WILDCARD_WORD = Word("*")

    def __init__(self, merge_ranges=False, add_head=" "):
        self.merge_ranges = merge_ranges
        self.add_head = add_head
        super().__init__(track_parents=True)

    def _get_node_bound_side(self, node):
        """Given a Range node, returns the bound side of the range (either 'high' or 'low').
        This guarantees that the other side is a wildcard. Returns None if the provided node is
        bound on both sides or bound on neither side.

        If the provided node is not a Range object, always returns None.
        """
        if isinstance(node, Range):
            if node.low == self.WILDCARD_WORD and node.high != self.WILDCARD_WORD:
                return 'high'
            elif node.low != self.WILDCARD_WORD and node.high == self.WILDCARD_WORD:
                return 'low'
        return None

    def visit_and_operation(self, node, context):
        if not self.merge_ranges:
            yield from self.generic_visit(node, context)
            return

        new_node = AndOperation(pos=node.pos, size=node.size, head=node.head, tail=node.tail)
        new_node.children = ()

        # We collect all ranges from a AND's children with the same bound side, and pop the
        # first one from the list when we encounter a different bound side. This allows us
        # to join multiple low bounds with multiple high bounds, or vice versa, without
        # requiring them to be in order (see class's docstring for an example)
        #
        # This works correctly because possible_ranges will be a list of open ranges from only one
        # side. Which side this is, is tracked in possible_ranges_bound_side. Any new node of the
        # other side will join into an already existing node, until the list is empty, and we may
        # reset possible_ranges_bound_side again.
        possible_ranges = []

        # possible_ranges_bound_side has no meaning if possible_ranges is empty, and may either
        # be None, or set to a previous value. When adding anything to possible_ranges, we make
        # sure that the direction is correct. All nodes in possible_ranges will be of the same
        # bound side.
        possible_ranges_bound_side = None

        for child in self.clone_children(node, new_node, context):
            # Determines the bound side of the range (high or low), meaning that the other side is
            # a wildcard.
            child_bound_side = self._get_node_bound_side(child)

            if child_bound_side is not None:
                # If we have encountered a Range with only one bound side, we may be able to join it
                if not possible_ranges or possible_ranges_bound_side == child_bound_side:
                    # We have not yet encountered any unbound ranges, or the encountered unbound
                    # ranges are of this node's direction, so store a pointer to this child and we
                    # may be able to join a future range. The flow will continue.
                    possible_ranges.append(child)
                    # Store the direction in the ranges. This assignment only has meaning if
                    # possible_ranges was empty, but the if statement prevents flipping it otherwise
                    possible_ranges_bound_side = child_bound_side
                else:
                    # There is at least one possible range to join to. We adjust the first range
                    # encountered of the different side
                    #
                    # We adjust the first enounter of the range, and do not add the joined one, to
                    # ensure the correct ordering of the different nodes (it is more intuitive to
                    # add it to the first one)
                    joining_child = possible_ranges.pop(0)
                    if child_bound_side == 'low':
                        joining_child.low = child.low
                        joining_child.include_low = child.include_low
                    else:
                        joining_child.high = child.high
                        joining_child.include_high = child.include_high
                    # we should not need to adjust the head/tail because those would be included
                    # from the child range we are joining from

                    # do not add the child we have just merged from
                    continue

            # Could not join the node (either because child_bound_side is None, or it has nothing to
            # join to)
            new_node.children += (child, )

        yield new_node

    def _visit_from_to(self, node, context, bound_side):
        # We omit the child directly for now, as we cannot assign it yet. We add the
        # wildcard element though, and the child is None until we have cloned it.
        if bound_side == 'low':
            args = [None, self.WILDCARD_WORD.clone_item(), node.include, True]
        else:
            args = [self.WILDCARD_WORD.clone_item(), None, True, node.include]

        new_node = Range(
            *args,
            pos=node.pos, size=node.size, head=node.head, tail=node.tail,
        )

        # this forces that we clone the children like we should, but we only expect
        # one return value, so we can simply assign it directly
        child = tuple(self.clone_children(node, new_node, context))[0]
        setattr(new_node, bound_side, child)

        new_node.low.tail += self.add_head
        new_node.high.head += self.add_head
        yield new_node

    def visit_from(self, node, context):
        yield from self._visit_from_to(node, context, 'low')

    def visit_to(self, node, context):
        yield from self._visit_from_to(node, context, 'high')

    def __call__(self, tree):
        return self.visit(tree)


def normalize_nested_fields_specs(nested_fields):
    """normalize nested_fields specification to only have nested dicts

    :param dict nested_fields:  dict contains fields that are nested in ES
        each nested fields contains either
        a dict of nested fields
        (if some of them are also nested)
        or a list of nesdted fields (this is for commodity)


    ::
        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     normalize_nested_fields_specs(
        ...         {"author" : {"books": ["name", "ref"], "firstname" : None }}),
        ...     {"author" : {"books": {"name": {}, "ref": {}}, "firstname" : {} }})
    """
    if nested_fields is None:
        return {}
    elif isinstance(nested_fields, dict):
        return {k: normalize_nested_fields_specs(v) for k, v in nested_fields.items()}
    else:
        # should be an iterable, transform to dict
        return {sub: {} for sub in nested_fields}


def _flatten_fields_specs(object_fields):
    if not object_fields:
        return [[]]  # parent is a single field
    elif isinstance(object_fields, dict):
        return [
            [k] + v2
            for k, v in object_fields.items()
            for v2 in _flatten_fields_specs(v)]
    else:  # iterable
        return [[k] for k in object_fields]


def flatten_nested_fields_specs(nested_fields):
    """normalize object_fields specification to only have a simple set

    :param dict nested_fields:  contains fields that are object in ES
        has a serie of nested dict.
        List are accepted as well for concisness.

    ::
        >>> from unittest import TestCase
        >>> flatten_nested_fields_specs(None)
        set()
        >>> TestCase().assertEqual(
        ...     flatten_nested_fields_specs(["author.name", "book.title"]),
        ...     set(["author.name", "book.title"]))
        >>> TestCase().assertEqual(
        ...     flatten_nested_fields_specs(
        ...         {"book" : { "author": ["firstname", "lastname"], "title" : None }}),
        ...     set(["book.author.firstname", "book.author.lastname", "book.title"]))
    """
    if isinstance(nested_fields, dict):
        return set(".".join(v) for v in _flatten_fields_specs(nested_fields))
    elif nested_fields is None:
        return set([])
    else:
        return set(nested_fields)


def normalize_object_fields_specs(object_fields):
    """normalize object_fields specification to only have a simple set

    :param dict object_fields:  contains fields that are object in ES
        has a serie of nested dict.
        List are accepted as well for concisness.
        None, which means no spec, is returned as is.

    ::
        >>> from unittest import TestCase
        >>> normalize_object_fields_specs(None) is None
        True
        >>> TestCase().assertEqual(
        ...     normalize_object_fields_specs(["author.name", "book.title"]),
        ...     set(["author.name", "book.title"]))
        >>> TestCase().assertEqual(
        ...     normalize_object_fields_specs(
        ...         {"book" : { "author": ["firstname", "lastname"], "title" : None }}),
        ...     set(["book.author.firstname", "book.author.lastname", "book.title"]))
    """
    if object_fields is None:
        return None
    if isinstance(object_fields, dict):
        return set(".".join(v) for v in _flatten_fields_specs(object_fields))
    else:
        return set(object_fields)



# -*- coding: utf-8 -*-
import collections
from unittest import TestCase

from luqum.head_tail import HeadTailLexer, HeadTailManager, TokenValue
from luqum.parser import parser
from luqum.tree import (
    AndOperation, Boost, FieldGroup, Fuzzy, Group, Item, Not,
    OrOperation, Phrase, Plus, Prohibit, Proximity,
    Range, Regex, SearchField, UnknownOperation, Word,
)


class Dummy:
    def __str__(self):
        return "silly me"


class TokenValueTestCase(TestCase):

    def test_tokenvalue_str(self):
        self.assertEqual(repr(TokenValue("")), "TokenValue()")
        self.assertEqual(str(TokenValue("")), "")
        self.assertEqual(repr(TokenValue("foo")), "TokenValue(foo)")
        self.assertEqual(str(TokenValue("foo")), "foo")
        self.assertEqual(str(TokenValue(None)), "")
        self.assertEqual(str(TokenValue(Dummy())), "silly me")

    def test_tokenvalue_head_tail_pos(self):
        t = TokenValue("foo")
        self.assertEqual(t.pos, None)
        self.assertEqual(t.head, "")
        self.assertEqual(t.tail, "")


TokenMock = collections.namedtuple("TokenMock", "type value lexpos lexer")


class LexerMock:
    pass


def token_factory():
    lexer = LexerMock()

    def create_token(type_, value, lexpos):
        return TokenMock(type_, value, lexpos, lexer)

    return create_token


class HeadTailLexerTestCase(TestCase):

    handle = HeadTailLexer.handle

    def test_separator_first_in_head(self):
        for value in (TokenValue("test"), Item()):
            create_token = token_factory()
            self.handle(create_token("SEPARATOR", "\t", 0), "\t")
            token = create_token("OTHER", value, 1)
            self.handle(token, "test")
            self.assertEqual(token.value.head, "\t")
            self.assertEqual(token.value.tail, "")
            self.assertEqual(token.value.pos, 1)
            self.assertEqual(token.value.size, 4)

    def test_token_first_pos(self):
        for value in (TokenValue("test"), Item()):
            create_token = token_factory()
            token = create_token("OTHER", value, 0)
            self.handle(token, "test")
            self.assertEqual(token.value.head, "")
            self.assertEqual(token.value.tail, "")
            self.assertEqual(token.value.pos, 0)
            self.assertEqual(token.value.size, 4)

    def test_simple_token_tail(self):
        create_token = token_factory()
        a = create_token("OTHER", TokenValue("a"), 0)
        b = create_token("OTHER", Item(), 3)
        c = create_token("OTHER", TokenValue("c"), 5)
        self.handle(a, "a")
        self.handle(create_token("SEPARATOR", " \t", 1), "\t")
        self.handle(b, "")
        self.handle(create_token("SEPARATOR", "\r", 4), "\r")
        self.handle(c, "c")
        self.assertEqual(a.value.head, "")
        self.assertEqual(a.value.tail, " \t")
        self.assertEqual(a.value.pos, 0)
        self.assertEqual(a.value.size, 1)
        self.assertEqual(b.value.head, "")
        self.assertEqual(b.value.tail, "\r")
        self.assertEqual(b.value.pos, 3)
        self.assertEqual(b.value.size, 0)
        self.assertEqual(c.value.head, "")
        self.assertEqual(c.value.tail, "")
        self.assertEqual(c.value.pos, 5)
        self.assertEqual(c.value.size, 1)

    def test_separator_last_elt_none(self):
        # this is a robustness test
        for value in (TokenValue("test"), Item()):
            create_token = token_factory()
            self.handle(create_token("SEPARATOR", "\t", 0), "\t")
            self.handle(create_token("SEPARATOR", "\n", 1), "\n")
            token = create_token("OTHER", value, 2)
            self.handle(token, "test")
            self.assertEqual(token.value.head, "\t")
            self.assertEqual(token.value.tail, "")
            self.assertEqual(token.value.pos, 2)
            self.assertEqual(token.value.size, 4)

    def test_tail_at_end(self):
        create_token = token_factory()
        a = create_token("OTHER", TokenValue("a"), 0)
        b = create_token("OTHER", Item(), 3)
        self.handle(a, "a")
        self.handle(create_token("SEPARATOR", " \t", 1), "\t")
        self.handle(b, "")
        # finish with a separator
        self.handle(create_token("SEPARATOR", "\r", 4), "\r")

        self.assertEqual(a.value.head, "")
        self.assertEqual(a.value.tail, " \t")
        self.assertEqual(a.value.pos, 0)
        self.assertEqual(a.value.size, 1)
        self.assertEqual(b.value.head, "")
        self.assertEqual(b.value.tail, "\r")
        self.assertEqual(b.value.pos, 3)
        self.assertEqual(b.value.size, 0)


class HeadTailManagerTestCase(TestCase):

    headtail = HeadTailManager()

    def test_pos(self):
        p = [Item(), Item(pos=4, size=3)]
        self.headtail.pos(p)
        self.assertEqual(p[0].pos, 4)
        self.assertEqual(p[0].size, 3)

    def test_pos_without_head_transfer(self):
        p = [Item(), Item(pos=4, size=3, head="\r\n")]
        self.headtail.pos(p, head_transfer=False)
        self.assertEqual(p[0].pos, 2)
        self.assertEqual(p[0].size, 5)

    def test_pos_with_head_transfer(self):
        p = [Item(), Item(pos=4, size=3, head="\r\n")]
        self.headtail.pos(p, head_transfer=True)
        self.assertEqual(p[0].pos, 4)
        self.assertEqual(p[0].size, 3)

    def test_pos_none(self):
        p = [Item(), Item(pos=None, size=None)]
        self.headtail.pos(p)
        self.assertEqual(p[0].pos, None)
        self.assertEqual(p[0].size, 0)

    def test_binary_operation(self):
        p = [
            Item(),
            Item(pos=1, size=3, head="\t", tail="\n"),
            Item(pos=4, size=4, head="\r", tail="  "),
        ]
        self.headtail.binary_operation(p, op_tail="")
        self.assertEqual(p[0].pos, 0)
        self.assertEqual(p[0].size, 12)
        self.headtail.binary_operation(p, op_tail="  ")
        self.assertEqual(p[0].pos, 0)
        self.assertEqual(p[0].size, 10)

    def test_simple_term(self):
        p = [Item(), Item(pos=4, size=3, head="\t", tail="\r")]
        self.headtail.simple_term(p)
        self.assertEqual(p[0].head, "\t")
        self.assertEqual(p[0].tail, "\r")
        self.assertEqual(p[0].pos, 4)
        self.assertEqual(p[0].size, 3)

    def test_unary(self):
        p = [
            Item(),
            Item(head="\t", tail="\r", pos=3, size=3),
            Item(head="\n", tail="  ", pos=5, size=5),
        ]
        self.headtail.unary(p)
        self.assertEqual(p[0].head, "\t")
        self.assertEqual(p[0].tail, "")
        self.assertEqual(p[0].pos, 3)
        self.assertEqual(p[0].size, 12)
        self.assertEqual(p[2].head, "\r\n")
        self.assertEqual(p[2].tail, "  ")
        self.assertEqual(p[2].pos, 5)
        self.assertEqual(p[2].size, 5)

    def test_post_unary(self):
        p = [
            Item(),
            Item(head="\t", tail="\r", pos=3, size=3),
            Item(head="\n", tail="  ", pos=5, size=5),
        ]
        self.headtail.post_unary(p)
        self.assertEqual(p[0].head, "")
        self.assertEqual(p[0].tail, "  ")
        self.assertEqual(p[0].pos, 2)
        self.assertEqual(p[0].size, 11)
        self.assertEqual(p[1].head, "\t")
        self.assertEqual(p[1].tail, "\r\n")
        self.assertEqual(p[1].pos, 3)
        self.assertEqual(p[1].pos, 3)

    def test_paren(self):
        p = [
            Item(),  # result
            Item(head="\t", tail="\r", pos=3, size=1),  # (
            Item(head="\n", tail="  ", pos=5, size=3),  # expr
            Item(head="\n\n", tail="\t\t", pos=7, size=1),  # )
        ]
        self.headtail.paren(p)
        # result
        self.assertEqual(p[0].head, "\t")
        self.assertEqual(p[0].tail, "\t\t")
        self.assertEqual(p[0].pos, 3)
        self.assertEqual(p[0].size, 11)
        # expr
        self.assertEqual(p[2].head, "\r\n")
        self.assertEqual(p[2].tail, "  \n\n")
        self.assertEqual(p[2].pos, 5)
        self.assertEqual(p[2].size, 3)

    def test_range(self):
        p = [
            Item(),  # result
            Item(head="\t", tail="\r", pos=3, size=1),  # [
            Item(head="\n", tail="  ", pos=5, size=3),  # expr1
            Item(head="\n\n", tail="\t\t", pos=7, size=2),  # TO
            Item(head="\r\r", tail=" \t ", pos=9, size=5),  # expr2
            Item(head=" \r ", tail=" \n ", pos=12, size=1),  # ]
        ]
        self.headtail.range(p)
        # result
        self.assertEqual(p[0].head, "\t")
        self.assertEqual(p[0].tail, " \n ")
        self.assertEqual(p[0].pos, 3)
        self.assertEqual(p[0].size, 28)
        # expr1
        self.assertEqual(p[2].head, "\r\n")
        self.assertEqual(p[2].tail, "  \n\n")
        self.assertEqual(p[2].pos, 5)
        self.assertEqual(p[2].size, 3)
        # expr2
        self.assertEqual(p[4].head, "\t\t\r\r")
        self.assertEqual(p[4].tail, " \t  \r ")
        self.assertEqual(p[4].pos, 9)
        self.assertEqual(p[4].size, 5)

    def test_search_field(self):
        p = [
            Item(),  # result
            Item(head="\t", tail="\r", pos=3, size=3),  # name
            Item(head="\n", tail="  ", pos=5, size=1),  # :
            Item(head="\n\n", tail="\t\t", pos=7, size=5),  # expr
        ]
        self.headtail.search_field(p)
        # result
        self.assertEqual(p[0].head, "\t")
        self.assertEqual(p[0].tail, "")
        self.assertEqual(p[0].pos, 3)
        self.assertEqual(p[0].size, 17)
        # expr
        self.assertEqual(p[3].head, "  \n\n")
        self.assertEqual(p[3].tail, "\t\t")
        self.assertEqual(p[3].pos, 7)


class IntegrationTestCase(TestCase):

    def test_word(self):
        tree = parser.parse("\tfoo\r")
        self.assertEqual(tree, Word("foo"))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "\r")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 3)
        self.assertEqual(str(tree), "foo")
        self.assertEqual(tree.__str__(head_tail=True), "\tfoo\r")

    def test_phrase(self):
        tree = parser.parse('\t"foo  bar"\r')
        self.assertEqual(tree, Phrase('"foo  bar"'))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "\r")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 10)
        self.assertEqual(str(tree), '"foo  bar"')
        self.assertEqual(tree.__str__(head_tail=True), '\t"foo  bar"\r')

    def test_regex(self):
        tree = parser.parse('\t/foo/\r')
        self.assertEqual(tree, Regex('/foo/'))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "\r")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 5)
        self.assertEqual(str(tree), '/foo/')
        self.assertEqual(tree.__str__(head_tail=True), '\t/foo/\r')

    def test_term_to(self):
        tree = parser.parse("\tTO\r")
        self.assertEqual(tree, Word("TO"))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "\r")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 2)
        self.assertEqual(str(tree), "TO")
        self.assertEqual(tree.__str__(head_tail=True), "\tTO\r")

    def test_unknown_operator(self):
        tree = parser.parse("\tfoo\nbar\r")
        self.assertEqual(tree, UnknownOperation(Word("foo"), Word("bar")))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 9)
        foo, bar = tree.children
        self.assertEqual(foo.head, "\t")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 3)
        self.assertEqual(bar.head, "")
        self.assertEqual(bar.tail, "\r")
        self.assertEqual(bar.pos, 5)
        self.assertEqual(bar.size, 3)
        self.assertEqual(str(tree), "\tfoo\nbar\r")
        self.assertEqual(tree.__str__(head_tail=True), "\tfoo\nbar\r")

    def test_or_operator(self):
        tree = parser.parse("\tfoo\nOR  bar\rOR\t\nbaz\r\r")
        self.assertEqual(tree, OrOperation(Word("foo"), Word("bar"), Word("baz")))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 22)
        foo, bar, baz = tree.children
        self.assertEqual(foo.head, "\t")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 3)
        self.assertEqual(bar.head, "  ")
        self.assertEqual(bar.tail, "\r")
        self.assertEqual(bar.pos, 9)
        self.assertEqual(bar.size, 3)
        self.assertEqual(baz.head, "\t\n")
        self.assertEqual(baz.tail, "\r\r")
        self.assertEqual(baz.pos, 17)
        self.assertEqual(baz.size, 3)
        self.assertEqual(str(tree), "\tfoo\nOR  bar\rOR\t\nbaz\r\r")
        self.assertEqual(tree.__str__(head_tail=True), "\tfoo\nOR  bar\rOR\t\nbaz\r\r")

    def test_and_operator(self):
        tree = parser.parse("\tfoo\nAND  bar\rAND\t\nbaz\r\r")
        self.assertEqual(tree, AndOperation(Word("foo"), Word("bar"), Word("baz")))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 24)
        foo, bar, baz = tree.children
        self.assertEqual(foo.head, "\t")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 3)
        self.assertEqual(bar.head, "  ")
        self.assertEqual(bar.tail, "\r")
        self.assertEqual(bar.pos, 10)
        self.assertEqual(bar.size, 3)
        self.assertEqual(baz.head, "\t\n")
        self.assertEqual(baz.tail, "\r\r")
        self.assertEqual(baz.pos, 19)
        self.assertEqual(baz.size, 3)
        self.assertEqual(str(tree), "\tfoo\nAND  bar\rAND\t\nbaz\r\r")
        self.assertEqual(tree.__str__(head_tail=True), "\tfoo\nAND  bar\rAND\t\nbaz\r\r")

    def test_plus(self):
        tree = parser.parse("\t+\rfoo\n")
        self.assertEqual(tree, Plus(Word("foo")))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 6)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 3)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "+\rfoo\n")
        self.assertEqual(tree.__str__(head_tail=True), "\t+\rfoo\n")

    def test_minus(self):
        tree = parser.parse("\t-\rfoo\n")
        self.assertEqual(tree, Prohibit(Word("foo")))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 6)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 3)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "-\rfoo\n")
        self.assertEqual(tree.__str__(head_tail=True), "\t-\rfoo\n")

    def test_not(self):
        tree = parser.parse("\tNOT\rfoo\n")
        self.assertEqual(tree, Not(Word("foo")))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 8)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 5)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "NOT\rfoo\n")
        self.assertEqual(tree.__str__(head_tail=True), "\tNOT\rfoo\n")

    def test_group(self):
        tree = parser.parse("\t(\rfoo  )\n")
        self.assertEqual(tree, Group(Word("foo")))
        self.assertEqual(tree.head, "\t")
        self.assertEqual(tree.tail, "\n")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 8)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "  ")
        self.assertEqual(foo.pos, 3)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "(\rfoo  )")
        self.assertEqual(tree.__str__(head_tail=True), "\t(\rfoo  )\n")

    def test_search_field(self):
        # FIXME handle space between field name and ':' ?
        tree = parser.parse("\rfoo:\tbar\n")
        self.assertEqual(tree, SearchField("foo", Word("bar")))
        self.assertEqual(tree.head, "\r")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 9)
        bar, = tree.children
        self.assertEqual(bar.head, "\t")
        self.assertEqual(bar.tail, "\n")
        self.assertEqual(bar.pos, 6)
        self.assertEqual(bar.size, 3)
        self.assertEqual(str(tree), "foo:\tbar\n")
        self.assertEqual(tree.__str__(head_tail=True), "\rfoo:\tbar\n")

    def test_field_group(self):
        tree = parser.parse("\rfoo:\t(  bar\n)\t\n")
        self.assertEqual(tree, SearchField("foo", FieldGroup(Word("bar"))))
        self.assertEqual(tree.head, "\r")
        self.assertEqual(tree.tail, "")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 15)
        group, = tree.children
        self.assertEqual(group.head, "\t")
        self.assertEqual(group.tail, "\t\n")
        self.assertEqual(group.pos, 6)
        self.assertEqual(group.size, 8)
        bar, = group.children
        self.assertEqual(bar.head, "  ")
        self.assertEqual(bar.tail, "\n")
        self.assertEqual(bar.pos, 9)
        self.assertEqual(bar.size, 3)
        self.assertEqual(str(tree), "foo:\t(  bar\n)\t\n")
        self.assertEqual(tree.__str__(head_tail=True), "\rfoo:\t(  bar\n)\t\n")

    def test_range(self):
        tree = parser.parse("\r{\tfoo\nTO  bar\r\n]\t\t")
        self.assertEqual(tree, Range(Word("foo"), Word("bar"), include_low=False))
        self.assertEqual(tree.head, "\r")
        self.assertEqual(tree.tail, "\t\t")
        self.assertEqual(tree.pos, 1)
        self.assertEqual(tree.size, 16)
        foo, bar = tree.children
        self.assertEqual(foo.head, "\t")
        self.assertEqual(foo.tail, "\n")
        self.assertEqual(foo.pos, 3)
        self.assertEqual(foo.size, 3)
        self.assertEqual(bar.head, "  ")
        self.assertEqual(bar.tail, "\r\n")
        self.assertEqual(bar.pos, 11)
        self.assertEqual(bar.size, 3)
        self.assertEqual(str(tree), "{\tfoo\nTO  bar\r\n]")
        self.assertEqual(tree.__str__(head_tail=True), "\r{\tfoo\nTO  bar\r\n]\t\t")

    def test_boosting(self):
        tree = parser.parse("\rfoo\t^2\n")
        self.assertEqual(tree, Boost(Word("foo"), 2))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "\n")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 7)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\t")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "\rfoo\t^2")
        self.assertEqual(tree.__str__(head_tail=True), "\rfoo\t^2\n")

    def test_fuzzy(self):
        tree = parser.parse("\rfoo\t~2\n")
        self.assertEqual(tree, Fuzzy(Word("foo"), 2))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "\n")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 7)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\t")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 3)
        self.assertEqual(str(tree), "\rfoo\t~2")
        self.assertEqual(tree.__str__(head_tail=True), "\rfoo\t~2\n")

    def test_proximity(self):
        tree = parser.parse('\r"foo"\t~2\n')
        self.assertEqual(tree, Proximity(Phrase('"foo"'), 2))
        self.assertEqual(tree.head, "")
        self.assertEqual(tree.tail, "\n")
        self.assertEqual(tree.pos, 0)
        self.assertEqual(tree.size, 9)
        foo, = tree.children
        self.assertEqual(foo.head, "\r")
        self.assertEqual(foo.tail, "\t")
        self.assertEqual(foo.pos, 1)
        self.assertEqual(foo.size, 5)
        self.assertEqual(str(tree), '\r"foo"\t~2')
        self.assertEqual(tree.__str__(head_tail=True), '\r"foo"\t~2\n')

    def test_complex(self):
        # the scope of head / tail management is to be able to keep original structure
        # event after tree transformation or so
        query = "\rfoo AND bar  \nAND \t(\rbaz OR    spam\rOR ham\t\t)\r"
        tree = parser.parse(query)
        self.assertEqual(str(tree), query)
        self.assertEqual(tree.__str__(head_tail=True), query)

    def test_head_on_topmost(self):
        # if the head and tail is on topmost element, the str alone will strip
        query = "\r(foo AND bar  \nAND \t(\rbaz OR    spam\rOR ham\t\t))\r"
        tree = parser.parse(query)
        self.assertEqual(str(tree), query.strip())
        self.assertEqual(tree.__str__(head_tail=True), query)



# -*- coding: utf-8 -*-
from unittest import TestCase

from luqum.tree import (
    SearchField, FieldGroup, Group, Word, Phrase, Proximity, Fuzzy, Range,
    Not, AndOperation, OrOperation, Plus, UnknownOperation)
from luqum.auto_head_tail import auto_head_tail


class AutoHeadTailTestCase(TestCase):

    def test_or_operation(self):
        tree = OrOperation(Word("foo"), Word("bar"), Word("baz"))
        self.assertEqual(str(tree), "fooORbarORbaz")
        self.assertEqual(str(auto_head_tail(tree)), "foo OR bar OR baz")

    def test_and_operation(self):
        tree = AndOperation(Word("foo"), Word("bar"), Word("baz"))
        self.assertEqual(str(tree), "fooANDbarANDbaz")
        self.assertEqual(str(auto_head_tail(tree)), "foo AND bar AND baz")

    def test_unknown_operation(self):
        tree = UnknownOperation(Word("foo"), Word("bar"), Word("baz"))
        self.assertEqual(str(tree), "foobarbaz")
        self.assertEqual(str(auto_head_tail(tree)), "foo bar baz")

    def test_range(self):
        tree = Range(Word("foo"), Word("bar"))
        self.assertEqual(str(tree), "[fooTObar]")
        self.assertEqual(str(auto_head_tail(tree)), "[foo TO bar]")

    def test_not(self):
        tree = Not(Word("foo"))
        self.assertEqual(str(tree), "NOTfoo")
        self.assertEqual(str(auto_head_tail(tree)), "NOT foo")

    def test_complex(self):
        tree = Group(
            OrOperation(
                SearchField(
                    "foo",
                    FieldGroup(UnknownOperation(Word("bar"), Range(Word("baz"), Word("spam")))),
                ),
                Not(Proximity(Phrase('"ham ham"'), 2)),
                Plus(Fuzzy(Word("hammer"), 3)),
            )
        )
        self.assertEqual(str(tree), '(foo:(bar[bazTOspam])ORNOT"ham ham"~2OR+hammer~3)')
        self.assertEqual(
            str(auto_head_tail(tree)),
            '(foo:(bar [baz TO spam]) OR NOT "ham ham"~2 OR +hammer~3)',
        )
        # idem potent
        self.assertEqual(
            str(auto_head_tail(auto_head_tail(tree))),
            '(foo:(bar [baz TO spam]) OR NOT "ham ham"~2 OR +hammer~3)',
        )

    def test_auto_head_tail_no_change_to_existing(self):
        tree = AndOperation(
            Range(Word("foo", tail="\t"), Word("bar", head="\n"), tail="\r"),
            Not(Word("baz", head="\t\t"), head="\n\n", tail="\r\r"),
            Word("spam", head="\t\n"),
        )
        self.assertEqual(str(tree), "[foo\tTO\nbar]\rAND\n\nNOT\t\tbaz\r\rAND\t\nspam")
        self.assertEqual(
            str(auto_head_tail(tree)),
            "[foo\tTO\nbar]\rAND\n\nNOT\t\tbaz\r\rAND\t\nspam"
        )



import queue
import threading

import ply.lex as lex

from luqum.parser import parser
from luqum.thread import parse
from tests import alternative_lexer


def test_thread_parse():

    result_queue = queue.Queue()
    qs1 = """
        (title:"foo bar" AND body:"quick fox") OR title:fox AND
        (title:"foo bar" AND body:"quick fox") OR
        title:fox AND (title:"foo bar" AND body:"quick fox") OR
        title:fox AND (title:"foo bar" AND body:"quick fox") OR
        title:fox AND (title:"foo bar" AND body:"quick fox") OR title:fox
    """
    expected_tree = parser.parse(qs1)

    def run(q):
        parse(qs1)
        tree = parse(qs1)
        q.put(tree)

    # make concurrents calls
    threads = [threading.Thread(target=run, args=(result_queue,)) for i in range(100)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    assert result_queue.qsize() == 100
    for i in range(100):
        assert result_queue.get() == expected_tree


def test_thread_lex_global_state():
    """
    Last Lexer is used globally by default by the parser. If another library
    creates another lexer, it should not impact luqum.

    More info: [Multiple Parsers and
    Lexers](http://www.dabeaz.com/ply/ply.html#ply_nn37)
    """
    qs = '(title:"foo bar" AND body:"quick fox")'

    lex.lex(module=alternative_lexer)
    # if there is a "luqum.exceptions.ParseSyntaxError", the wrong lexer was
    # used.
    parse(qs)



from unittest import TestCase

from luqum.check import LuceneCheck, CheckNestedFields
from luqum.exceptions import NestedSearchFieldException, ObjectSearchFieldException
from luqum.parser import parser
from luqum.tree import (
    SearchField, FieldGroup, Group, Word, Phrase, Proximity, Fuzzy, Boost, Range,
    Not, AndOperation, OrOperation, Plus, Prohibit)


class TestCheck(TestCase):

    def test_check_ok(self):
        query = (
            AndOperation(
                SearchField(
                    "f",
                    FieldGroup(
                        AndOperation(
                            Boost(Proximity(Phrase('"foo bar"'), 4), "4.2"),
                            Prohibit(Range("100", "200"))))),
                Group(
                    OrOperation(
                        Fuzzy(Word("baz"), ".8"),
                        Plus(Word("fizz"))))))
        check = LuceneCheck()
        self.assertTrue(check(query))
        self.assertEqual(check.errors(query), [])
        check = LuceneCheck(zeal=1)
        self.assertTrue(check(query))
        self.assertEqual(check.errors(query), [])

    def test_bad_fieldgroup(self):
        check = LuceneCheck()
        query = FieldGroup(Word("foo"))
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("FieldGroup misuse", check.errors(query)[0])

        query = OrOperation(
            FieldGroup(Word("bar")),
            Word("foo"))
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("FieldGroup misuse", check.errors(query)[0])

    def test_bad_group(self):
        check = LuceneCheck()
        query = SearchField("f", Group(Word("foo")))
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 2)  # one for bad expr, one for misuse
        self.assertIn("Group misuse", "".join(check.errors(query)))

    def test_zealous_or_not_prohibit(self):
        query = (
            OrOperation(
                Prohibit(Word("foo")),
                Word("bar")))
        check_zealous = LuceneCheck(zeal=1)
        self.assertFalse(check_zealous(query))
        self.assertIn("inconsistent", check_zealous.errors(query)[0])
        check_easy_going = LuceneCheck()
        self.assertTrue(check_easy_going(query))

    def test_zealous_or_not(self):
        query = (
            OrOperation(
                Not(Word("foo")),
                Word("bar")))
        check_zealous = LuceneCheck(zeal=1)
        self.assertFalse(check_zealous(query))
        self.assertIn("inconsistent", check_zealous.errors(query)[0])
        check_easy_going = LuceneCheck()
        self.assertTrue(check_easy_going(query))

    def test_bad_field_name(self):
        check = LuceneCheck()
        query = SearchField("foo*", Word("bar"))
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("not a valid field name", check.errors(query)[0])

    def test_bad_field_expr(self):
        check = LuceneCheck()
        query = SearchField("foo", Prohibit(Word("bar")))
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("not valid", check.errors(query)[0])

    def test_word_space(self):
        check = LuceneCheck()
        query = Word("foo bar")
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("space", check.errors(query)[0])

    def test_invalid_characters_in_word_value(self):
        query = Word("foo/bar")
        # Passes if zeal == 0
        check = LuceneCheck()
        self.assertTrue(check(query))
        self.assertEqual(len(check.errors(query)), 0)
        # But not if zeal == 1
        check = LuceneCheck(zeal=1)
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("Invalid characters", check.errors(query)[0])

    def test_fuzzy_negative_degree(self):
        check = LuceneCheck()
        query = Fuzzy(Word("foo"), "-4.1")
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("invalid degree", check.errors(query)[0])

    def test_fuzzy_non_word(self):
        check = LuceneCheck()
        query = Fuzzy(Phrase('"foo bar"'), "2")
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("single term", check.errors(query)[0])

    def test_proximity_non_phrase(self):
        check = LuceneCheck()
        query = Proximity(Word("foo"), "2")
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 1)
        self.assertIn("phrase", check.errors(query)[0])

    def test_unknown_item_type(self):
        check = LuceneCheck()
        query = AndOperation("foo", 2)
        self.assertFalse(check(query))
        self.assertEqual(len(check.errors(query)), 2)
        self.assertIn("Unknown item type", check.errors(query)[0])
        self.assertIn("Unknown item type", check.errors(query)[1])


class CheckVisitorTestCase(TestCase):

    NESTED_FIELDS = {
        'author': {
            'firstname': {},
            'book': {
                'title': {},
                'format': {
                    'type': {}
                }
            }
        },
        'collection.keywords': {  # nested field inside an object field
            'key': {},
            'more_info.linked': {  # again nested field inside an object field
                'key': {}
            },
        },
    }

    OBJECT_FIELDS = [
        'author.birth.city',
        'collection.title', 'collection.ref', 'collection.keywords.more_info.revision']

    SUB_FIELDS = [
        'foo.english',
        'author.book.title.raw',
    ]

    def setUp(self):
        self.checker = CheckNestedFields(nested_fields=self.NESTED_FIELDS)
        self.strict_checker = CheckNestedFields(
            nested_fields=self.NESTED_FIELDS,
            object_fields=self.OBJECT_FIELDS,
            sub_fields=self.SUB_FIELDS,
        )

    def test_correct_nested_lucene_query_column_not_raise(self):
        tree = parser.parse('author:book:title:"foo" AND '
                            'author:book:format:type: "pdf"')
        self.strict_checker(tree)

    def test_correct_object_lucene_query_column_not_raise(self):
        tree = parser.parse('author:birth:city:"foo" AND '
                            'collection:(ref:"foo" AND title:"bar")')
        self.strict_checker(tree)
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_correct_subfield_lucene_query_column_not_raises(self):
        tree = parser.parse('foo:english:"foo" AND '
                            'author:book:title:raw:"pdf"')
        self.strict_checker(tree)

    def test_correct_nested_lucene_query_with_point_not_raise(self):
        tree = parser.parse('author.book.title:"foo" AND '
                            'author.book.format.type:"pdf"')
        self.strict_checker(tree)
        self.assertIsNotNone(tree)

    def test_correct_object_lucene_query_with_point_not_raise(self):
        tree = parser.parse('author.birth.city:"foo" AND '
                            'collection.ref:"foo"')
        self.strict_checker(tree)
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_correct_subfield_lucene_query_with_point_not_raises(self):
        tree = parser.parse('foo.english:"foo" AND '
                            'author.book.title.raw:"pdf"')
        self.strict_checker(tree)

    def test_correct_object_mix_do_not_raise(self):
        tree = parser.parse('author:(birth.city:"foo" AND book.title:"bar")')
        self.strict_checker(tree)
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_incorrect_nested_lucene_query_column_raise(self):
        tree = parser.parse('author:gender:"Mr" AND '
                            'author:book:format:type:"pdf"')
        with self.assertRaises(ObjectSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('author.gender', str(e.exception))

    def test_incorrect_nested_lucene_query_with_point_raise(self):
        tree = parser.parse('author.gender:"Mr" AND '
                            'author.book.format.type:"pdf"')
        with self.assertRaises(ObjectSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author.gender"', str(e.exception))

    def test_correct_nested_lucene_query_with_and_column_not_raise(self):
        tree = parser.parse(
            'author:(book.title:"foo" OR book.title:"bar")')
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_complex_subfield_not_raises(self):
        tree = parser.parse(
            'author:(book.title.raw:"foo" OR book.title.raw:"bar")')
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_simple_query_with_a_nested_field_should_raise(self):
        tree = parser.parse('author:"foo"')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author"', str(e.exception))

    def test_simple_query_with_a_multi_nested_field_should_raise(self):
        tree = parser.parse('author:book:"foo"')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author.book"', str(e.exception))

    def test_complex_query_with_a_multi_nested_field_should_raise(self):
        tree = parser.parse('author:test OR author.firstname:"Hugo"')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author"', str(e.exception))

    def test_complex_query_column_with_a_multi_nested_field_should_raise(self):
        tree = parser.parse('author:("test" AND firstname:Hugo)')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author"', str(e.exception))

    def test_complex_mix(self):
        tree = parser.parse(
            'collection:(title:"foo" AND keywords.more_info:(linked.key:"bar" revision:"test"))')
        self.strict_checker(tree)
        self.checker(tree)
        self.assertIsNotNone(tree)

    def test_complex_mix_raise(self):
        tree = parser.parse(
            'collection:(title:"foo" AND keywords.more_info:(linked:"bar" revision:"test"))')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"collection.keywords.more_info.linked"', str(e.exception))
        self.assertIsNotNone(tree)

    def test_incomplete_object_field_raise(self):
        tree = parser.parse('collection.keywords.more_info:"foo"')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"collection.keywords.more_info"', str(e.exception))

        tree = parser.parse('author:birth:"foo"')
        with self.assertRaises(NestedSearchFieldException) as e:
            self.strict_checker(tree)
        self.assertIn('"author.birth"', str(e.exception))
